#!/usr/bin/env python
# encoding: utf-8
"""
UtilizationSupplement.py

Copyright (c) NREL. All rights reserved.
"""

from math import atan2
import numpy as np
from wisdem.commonse.constants import eps
from wisdem.commonse.utilities import CubicSplineSegment, cubic_spline_eval, smooth_max, smooth_min, nodal2sectional, sectional2nodal
from openmdao.api import ExplicitComponent
from scipy.optimize import brentq, minimize_scalar

#-------------------------------------------------------------------------------
# Name:        UtilizationSupplement.py
# Purpose:     It contains functions to calculate utilizations for cylindric members,
#              for instance tower sections and monopile
#
# Author:      ANing/RRD/GBarter
#
# Created:     07/14/2015 - It is based on towerSupplement.py by ANing, 2012.
# Copyright:   (c) rdamiani 2015
# Licence:     <Apache 2015>
#-------------------------------------------------------------------------------



class GeometricConstraints(ExplicitComponent):
    """docstring for OtherConstraints"""
    def initialize(self):
        self.options.declare('nPoints')
        self.options.declare('diamFlag', default=True)
        
    def setup(self):
        nPoints       = self.options['nPoints']
        
        self.add_input('d', np.zeros(nPoints), units='m')
        self.add_input('t', np.zeros(nPoints-1), units='m')
        self.add_input('min_d_to_t', 120.0)
        self.add_input('max_taper', 0.4)

        self.add_output('weldability', np.zeros(nPoints-1))
        self.add_output('manufacturability', np.zeros(nPoints-1))
        self.add_output('slope', np.zeros(nPoints-2))

        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)


    def compute(self, inputs, outputs):
        diamFlag = self.options['diamFlag']

        d,_ = nodal2sectional(inputs['d'])
        t = inputs['t']

        # Check if the input was radii instead of diameters and convert if necessary
        if not diamFlag: d *= 2.0
        
        min_d_to_t = inputs['min_d_to_t']
        max_taper = inputs['max_taper']

        outputs['weldability'] = 1.0 - (d/t)/min_d_to_t
        d_ratio = d[1:]/d[:-1]
        manufacturability = np.minimum(d_ratio, 1.0/d_ratio) - max_taper
        outputs['manufacturability'] = np.r_[manufacturability, manufacturability[-1]]
        outputs['slope'] = d_ratio

    # def compute_partials(self, inputs, J):

    #     dw_dd = np.diag(-1.0/self.t/self.min_d_to_t)
    #     dw_dt = np.diag(self.d/self.t**2/self.min_d_to_t)

    #     dw = np.hstack([dw_dd, dw_dt])



    #     dm_dd = np.zeros_like(self.d)
    #     dm_dd[0] = self.d[-1]/self.d[0]**2
    #     dm_dd[-1] = -1.0/self.d[0]

    #     dm = np.hstack([dm_dd, np.zeros(len(self.t))])


def fatigue(M_DEL, N_DEL, d, t, m=4, DC=80.0, eta=1.265, stress_factor=1.0, weld_factor=True):
    """estimate fatigue damage for tower station

    Parmeters
    ---------
    M_DEL : array_like(float) (N*m)
        damage equivalent moment at tower section
    N_DEL : array_like(int)
        corresponding number of cycles in lifetime
    d : array_like(float) (m)
        tower diameter at section
    t : array_like(float) (m)
        tower shell thickness at section
    m : int
        slope of S/N curve
    DC : float (N/mm**2)
        some max stress from a standard
    eta : float
        safety factor
    stress_factor : float
        load_factor * stress_concentration_factor
    weld_factor : bool
        if True include an empirical weld factor

    Returns
    -------
    damage : float
        damage from Miner's rule for this tower section
    """


    # convert to mm
    dvec = np.array(d)*1e3
    tvec = np.array(t)*1e3

    t = sectional2nodal(t)
    nvec = len(d)
    damage = np.zeros(nvec)

    # initialize weld factor (added cubic spline around corner)
    if weld_factor:
        x1 = 24.0
        x2 = 26.0
        spline = CubicSplineSegment(x1, x2, 1.0, (25.0/x2)**0.25, 0.0, 25.0**0.25*-0.25*x2**-1.25)


    for i in range(nvec):

        d = dvec[i]
        t = tvec[i]

        # weld factor
        if not weld_factor or t <= x1:
            weld = 1.0
        elif t >= x2:
            weld = (25.0/t)**0.25
        else:
            weld = spline.eval(t)


        # stress
        r = d/2.0
        I = np.pi*r**3*t
        c = r
        sigma = M_DEL[i]*c/I * stress_factor * 1e3  # convert to N/mm^2

        # maximum allowed stress
        Smax = DC * weld / eta

        # number of cycles to failure
        Nf = (Smax/sigma)**m

        # number of cycles for this load
        N1 = 2e6  # TODO: where does this come from?
        N = N_DEL[i]/N1

        # damage
        damage[i] = N/Nf

    return damage


def vonMisesStressUtilization(axial_stress, hoop_stress, shear_stress, gamma, sigma_y):
    """combine stress for von Mises"""

    # von mises stress
    a = ((axial_stress + hoop_stress)/2.0)**2
    b = ((axial_stress - hoop_stress)/2.0)**2
    c = shear_stress**2
    von_mises = np.sqrt(a + 3.0*(b+c))

    # stress margin
    stress_utilization = gamma * von_mises / sigma_y

    return stress_utilization  # This must be <1 to pass

def hoopStress(d, t, q_dyn):
    r = d/2.0-t/2.0  # radius of cylinder middle surface
    return (-q_dyn * r / t)

def hoopStressEurocode(z, d, t, L_reinforced, q_dyn):
    """default method for computing hoop stress using Eurocode method
       GB 06/21/2018: Ansys comparisons for submerged case suggests this over-compensates for stiffener
                      I'm not even sure the Eurocode is implemented correctly here.  Suggest using the standard
                      hoop stress expression above or API's handling of ring stiffeners below.
    """

    r = d/2.0-t/2.0  # radius of cylinder middle surface
    omega = L_reinforced/np.sqrt(r*t)

    C_theta = 1.5  # clamped-clamped
    k_w = 0.46*(1.0 + 0.1*np.sqrt(C_theta/omega*r/t))
    kw = smooth_max(k_w, 0.65)
    kw = smooth_min(k_w, 1.0)
    Peq = k_w*q_dyn
    return hoopStress(d, t, Peq)


def bucklingGL(d, t, Fz, Myy, tower_height, E, sigma_y, gamma_f=1.2, gamma_b=1.1, gamma_g=1.1):

    # other factors
    alpha = 0.21  # buckling imperfection factor
    beta = 1.0  # bending coefficient
    sk_factor = 2.0  # fixed-free
    tower_height = tower_height * sk_factor

    # geometry
    A = np.pi * d * t
    I = np.pi * (d/2.0)**3 * t
    Wp = I / (d/2.0)

    # applied loads
    Nd = -Fz * gamma_g
    Md = Myy * gamma_f

    # plastic resistance
    Np = A * sigma_y / gamma_b
    Mp = Wp * sigma_y / gamma_b

    # factors
    Ne = np.pi**2 * (E * I) / (1.1 * tower_height**2)
    lambda_bar = np.sqrt(Np * gamma_b / Ne)
    phi = 0.5 * (1 + alpha*(lambda_bar - 0.2) + lambda_bar**2)
    kappa = np.ones_like(d)
    idx = lambda_bar > 0.2
    kappa[idx] = 1.0 / (phi[idx] + np.sqrt(phi[idx]**2 - lambda_bar[idx]**2))
    delta_n = 0.25*kappa*lambda_bar**2
    delta_n = np.minimum(delta_n, 0.1)

    GL_utilization = Nd/(kappa*Np) + beta*Md/Mp + delta_n  #this is utilization must be <1

    return GL_utilization





def shellBucklingEurocode(d, t, sigma_z, sigma_t, tau_zt, L_reinforced, E, sigma_y, gamma_f=1.2, gamma_b=1.1):
    """
    Estimate shell buckling utilization along tower.

    Arguments:
    npt - number of locations at each node at which stress is evaluated.
    sigma_z - axial stress at npt*node locations.  must be in order
                  [(node1_pts1-npt), (node2_pts1-npt), ...]
    sigma_t - azimuthal stress given at npt*node locations
    tau_zt - shear stress (z, theta) at npt*node locations
    E - modulus of elasticity
    sigma_y - yield stress
    L_reinforced - reinforcement length - structure is re-discretized with this spacing
    gamma_f - safety factor for stresses
    gamma_b - safety factor for buckling

    Returns:
    z
    EU_utilization: - array of shell buckling utilizations evaluted at (z[0] at npt locations, \n
                      z[0]+L_reinforced at npt locations, ...). \n
                      Each utilization must be < 1 to avoid failure.
    """

    n = len(d)
    EU_utilization = np.zeros(n)
    sigma_z_sh = np.zeros(n)
    sigma_t_sh = np.zeros(n)
    tau_zt_sh = np.zeros(n)

    for i in range(n):
        h = L_reinforced[i]

        r1 = d[i]/2.0 - t[i]/2.0
        r2 = d[i]/2.0 - t[i]/2.0

        sigma_z_shell = sigma_z[i]
        sigma_t_shell = sigma_t[i]
        tau_zt_shell = tau_zt[i]

        # TODO: the following is non-smooth, although in general its probably OK
        # change to magnitudes and add safety factor
        sigma_z_shell = gamma_f*abs(sigma_z_shell)
        sigma_t_shell = gamma_f*abs(sigma_t_shell)
        tau_zt_shell = gamma_f*abs(tau_zt_shell)

        EU_utilization[i] = _shellBucklingOneSection(h, r1, r2, t[i], gamma_b, sigma_z_shell, sigma_t_shell, tau_zt_shell, E[i], sigma_y[i])

        #make them into vectors
        sigma_z_sh[i]=sigma_z_shell
        sigma_t_sh[i]=sigma_t_shell
        tau_zt_sh[i]=tau_zt_shell

    return EU_utilization  # this is utilization must be <1




def _cxsmooth(omega, rovert):

    Cxb = 6.0  # clamped-clamped
    constant = 1 + 1.83/1.7 - 2.07/1.7**2

    ptL1 = 1.7-0.25
    ptR1 = 1.7+0.25

    ptL2 = 0.5*rovert - 1.0
    ptR2 = 0.5*rovert + 1.0

    ptL3 = (0.5+Cxb)*rovert - 1.0
    ptR3 = (0.5+Cxb)*rovert + 1.0


    if omega < ptL1:
        Cx = constant - 1.83/omega + 2.07/omega**2

    elif omega >= ptL1 and omega <= ptR1:

        fL = constant - 1.83/ptL1 + 2.07/ptL1**2
        fR = 1.0
        gL = 1.83/ptL1**2 - 4.14/ptL1**3
        gR = 0.0
        Cx = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        Cx = 1.0

    elif omega >= ptL2 and omega <= ptR2:

        fL = 1.0
        fR = 1 + 0.2/Cxb*(1-2.0*ptR2/rovert)
        gL = 0.0
        gR = -0.4/Cxb/rovert
        Cx = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    elif omega > ptR2 and omega < ptL3:
        Cx = 1 + 0.2/Cxb*(1-2.0*omega/rovert)

    elif omega >= ptL3 and omega <= ptR3:

        fL = 1 + 0.2/Cxb*(1-2.0*ptL3/rovert)
        fR = 0.6
        gL = -0.4/Cxb/rovert
        gR = 0.0
        Cx = cubic_spline_eval(ptL3, ptR3, fL, fR, gL, gR, omega)

    else:
        Cx = 0.6

    return Cx


def _sigmasmooth(omega, E, rovert):

    Ctheta = 1.5  # clamped-clamped

    ptL = 1.63*rovert*Ctheta - 1
    ptR = 1.63*rovert*Ctheta + 1

    if omega < 20.0*Ctheta:
        offset = (10.0/(20*Ctheta)**2 - 5/(20*Ctheta)**3)
        Cthetas = 1.5 + 10.0/omega**2 - 5/omega**3 - offset
        sigma = 0.92*E*Cthetas/omega/rovert

    elif omega >= 20.0*Ctheta and omega < ptL:

        sigma = 0.92*E*Ctheta/omega/rovert

    elif omega >= ptL and omega <= ptR:

        alpha1 = 0.92/1.63 - 2.03/1.63**4

        fL = 0.92*E*Ctheta/ptL/rovert
        fR = E*(1.0/rovert)**2*(alpha1 + 2.03*(Ctheta/ptR*rovert)**4)
        gL = -0.92*E*Ctheta/rovert/ptL**2
        gR = -E*(1.0/rovert)*2.03*4*(Ctheta/ptR*rovert)**3*Ctheta/ptR**2

        sigma = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, omega)

    else:

        alpha1 = 0.92/1.63 - 2.03/1.63**4
        sigma = E*(1.0/rovert)**2*(alpha1 + 2.03*(Ctheta/omega*rovert)**4)

    return sigma


def _tausmooth(omega, rovert):

    ptL1 = 9
    ptR1 = 11

    ptL2 = 8.7*rovert - 1
    ptR2 = 8.7*rovert + 1

    if omega < ptL1:
        C_tau = np.sqrt(1.0 + 42.0/omega**3 - 42.0/10**3)

    elif omega >= ptL1 and omega <= ptR1:
        fL = np.sqrt(1.0 + 42.0/ptL1**3 - 42.0/10**3)
        fR = 1.0
        gL = -63.0/ptL1**4/fL
        gR = 0.0
        C_tau = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        C_tau = 1.0

    elif omega >= ptL2 and omega <= ptR2:
        fL = 1.0
        fR = 1.0/3.0*np.sqrt(ptR2/rovert) + 1 - np.sqrt(8.7)/3
        gL = 0.0
        gR = 1.0/6/np.sqrt(ptR2*rovert)
        C_tau = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    else:
        C_tau = 1.0/3.0*np.sqrt(omega/rovert) + 1 - np.sqrt(8.7)/3

    return C_tau



def _shellBucklingOneSection(h, r1, r2, t, gamma_b, sigma_z, sigma_t, tau_zt, E, sigma_y):
    """
    Estimate shell buckling for one tapered cylindrical shell section.

    Arguments:
    h - height of conical section
    r1 - radius at bottom
    r2 - radius at top
    t - shell thickness
    E - modulus of elasticity
    sigma_y - yield stress
    gamma_b - buckling reduction safety factor
    sigma_z - axial stress component
    sigma_t - azimuthal stress component
    tau_zt - shear stress component (z, theta)

    Returns:
    EU_utilization, shell buckling utilization which must be < 1 to avoid failure

    """

    #NOTE: definition of r1, r2 switched from Eurocode document to be consistent with FEM.

    # ----- geometric parameters --------
    beta = atan2(r1-r2, h)
    L = h/np.cos(beta)

    # ------------- axial stress -------------
    # length parameter
    le = L
    re = 0.5*(r1+r2)/np.cos(beta)
    omega = le/np.sqrt(re*t)
    rovert = re/t

    # compute Cx
    Cx = _cxsmooth(omega, rovert)


    # if omega <= 1.7:
    #     Cx = 1.36 - 1.83/omega + 2.07/omega/omega
    # elif omega > 0.5*rovert:
    #     Cxb = 6.0  # clamped-clamped
    #     Cx = max(0.6, 1 + 0.2/Cxb*(1-2.0*omega/rovert))
    # else:
    #     Cx = 1.0

    # critical axial buckling stress
    sigma_z_Rcr = 0.605*E*Cx/rovert

    # compute buckling reduction factors
    lambda_z0 = 0.2
    beta_z = 0.6
    eta_z = 1.0
    Q = 25.0  # quality parameter - high
    lambda_z = np.sqrt(sigma_y/sigma_z_Rcr)
    delta_wk = 1.0/Q*np.sqrt(rovert)*t
    alpha_z = 0.62/(1 + 1.91*(delta_wk/t)**1.44)

    chi_z = _buckling_reduction_factor(alpha_z, beta_z, eta_z, lambda_z0, lambda_z)

    # design buckling stress
    sigma_z_Rk = chi_z*sigma_y
    sigma_z_Rd = sigma_z_Rk/gamma_b

    # ---------------- hoop stress ------------------

    # length parameter
    le = L
    re = 0.5*(r1+r2)/(np.cos(beta))
    omega = le/np.sqrt(re*t)
    rovert = re/t

    # Ctheta = 1.5  # clamped-clamped
    # CthetaS = 1.5 + 10.0/omega**2 - 5.0/omega**3

    # # critical hoop buckling stress
    # if (omega/Ctheta < 20.0):
    #     sigma_t_Rcr = 0.92*E*CthetaS/omega/rovert
    # elif (omega/Ctheta > 1.63*rovert):
    #     sigma_t_Rcr = E*(1.0/rovert)**2*(0.275 + 2.03*(Ctheta/omega*rovert)**4)
    # else:
    #     sigma_t_Rcr = 0.92*E*Ctheta/omega/rovert

    sigma_t_Rcr = np.maximum(eps, _sigmasmooth(omega, E, rovert))

    # buckling reduction factor
    alpha_t = 0.65  # high fabrication quality
    lambda_t0 = 0.4
    beta_t = 0.6
    eta_t = 1.0
    lambda_t = np.sqrt(sigma_y/sigma_t_Rcr)

    chi_theta = _buckling_reduction_factor(alpha_t, beta_t, eta_t, lambda_t0, lambda_t)

    sigma_t_Rk = chi_theta*sigma_y
    sigma_t_Rd = sigma_t_Rk/gamma_b

    # ----------------- shear stress ----------------------

    # length parameter
    le = h
    rho = np.sqrt((r1+r2)/(2.0*r2))
    re = (1.0 + rho - 1.0/rho)*r2*np.cos(beta)
    omega = le/np.sqrt(re*t)
    rovert = re/t

    # if (omega < 10):
    #     C_tau = np.sqrt(1.0 + 42.0/omega**3)
    # elif (omega > 8.7*rovert):
    #     C_tau = 1.0/3.0*np.sqrt(omega/rovert)
    # else:
    #     C_tau = 1.0
    C_tau = _tausmooth(omega, rovert)

    tau_zt_Rcr = 0.75*E*C_tau*np.sqrt(1.0/omega)/rovert

    # reduction factor
    alpha_tau = 0.65  # high fabrifaction quality
    beta_tau = 0.6
    lambda_tau0 = 0.4
    eta_tau = 1.0
    lambda_tau = np.sqrt(sigma_y/np.sqrt(3)/tau_zt_Rcr)

    chi_tau = _buckling_reduction_factor(alpha_tau, beta_tau, eta_tau, lambda_tau0, lambda_tau)

    tau_zt_Rk = chi_tau*sigma_y/np.sqrt(3)
    tau_zt_Rd = tau_zt_Rk/gamma_b

    # buckling interaction parameters

    k_z = 1.25 + 0.75*chi_z
    k_theta = 1.25 + 0.75*chi_theta
    k_tau = 1.75 + 0.25*chi_tau
    k_i = (chi_z*chi_theta)**2

    # shell buckling utilization

    utilization = \
        (sigma_z/sigma_z_Rd)**k_z + \
        (sigma_t/sigma_t_Rd)**k_theta - \
        k_i*(sigma_z*sigma_t/sigma_z_Rd/sigma_t_Rd) + \
        (tau_zt/tau_zt_Rd)**k_tau

    return utilization #this is utilization must be <1



def _buckling_reduction_factor(alpha, beta, eta, lambda_0, lambda_bar):
    """
    Computes a buckling reduction factor used in Eurocode shell buckling formula.
    """

    lambda_p = np.sqrt(alpha/(1.0-beta))

    ptL = 0.9*lambda_0
    ptR = 1.1*lambda_0

    if (lambda_bar < ptL):
        chi = 1.0

    elif lambda_bar >= ptL and lambda_bar <= ptR:  # cubic spline section

        fracR = (ptR-lambda_0)/(lambda_p-lambda_0)
        fL = 1.0
        fR = 1-beta*fracR**eta
        gL = 0.0
        gR = -beta*eta*fracR**(eta-1)/(lambda_p-lambda_0)

        chi = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, lambda_bar)

    elif lambda_bar > ptR and lambda_bar < lambda_p:
        chi = 1.0 - beta*((lambda_bar-lambda_0)/(lambda_p-lambda_0))**eta

    else:
        chi = alpha/lambda_bar**2



    # if (lambda_bar <= lambda_0):
    #     chi = 1.0
    # elif (lambda_bar >= lambda_p):
    #     chi = alpha/lambda_bar**2
    # else:
    #     chi = 1.0 - beta*((lambda_bar-lambda_0)/(lambda_p-lambda_0))**eta

    return chi




def _TBeamProperties(h_web, t_web, w_flange, t_flange):
    """Computes T-cross section area, CG, and moments of inertia
    See: http://www.amesweb.info/SectionalPropertiesTabs/SectionalPropertiesTbeam.aspx

    INPUTS:
    ----------
    h_web    : float (scalar/vector),  web (T-base) height
    t_web    : float (scalar/vector),  web (T-base) thickness
    w_flange : float (scalar/vector),  flange (T-top) width/height
    t_flange : float (scalar/vector),  flange (T-top) thickness

    OUTPUTS:
    -------
    area : float (scalar/vector),  T-cross sectional area
    y_cg : float (scalar/vector),  Position of CG along y-axis (extending from base up through the T)
    Ixx  : float (scalar/vector),  Moment of intertia around axis parallel to flange, through y_cg
    Iyy  : float (scalar/vector),  Moment of intertia around y-axis
    """
    # Area of T cross section is sum of the two rectangles
    area_web    = h_web * t_web
    area_flange = w_flange * t_flange
    area        = area_web + area_flange
    # Y-position of the center of mass (Yna) measured from the base
    y_cg = ( (h_web + 0.5*t_flange)*area_flange + 0.5*h_web*area_web ) / area
    # Moments of inertia: y-axis runs through base (spinning top),
    # x-axis runs parallel to flange through cg
    Iyy =  (area_web*t_web**2 + area_flange*w_flange**2    ) / 12.0
    Ixx = ((area_web*h_web**2 + area_flange*t_flange**2) / 12.0 +
           area_web*(y_cg - 0.5*h_web)**2 +
           area_flange*(h_web + 0.5*t_flange - y_cg)**2 )
    return area, y_cg, Ixx, Iyy


def _IBeamProperties(h_web, t_web, w_flange, t_flange, w_base, t_base):
    """Computes uneven I-cross section area, CG
    See: http://www.amesweb.info/SectionalPropertiesTabs/SectionalPropertiesTbeam.aspx

    INPUTS:
    ----------
    h_web    : float (scalar/vector),  web (I-stem) height
    t_web    : float (scalar/vector),  web (I-stem) thickness
    w_flange : float (scalar/vector),  flange (I-top) width/height
    t_flange : float (scalar/vector),  flange (I-top) thickness
    w_base   : float (scalar/vector),  base (I-bottom) width/height
    t_base   : float (scalar/vector),  base (I-bottom) thickness

    OUTPUTS:
    -------
    area : float (scalar/vector),  T-cross sectional area
    y_cg : float (scalar/vector),  Position of CG along y-axis (extending from base up through the T)
    """
    # Area of T cross section is sum of the two rectangles
    area_web    = h_web * t_web
    area_flange = w_flange * t_flange
    area_base   = w_base * t_base
    area        = area_web + area_flange + area_base
    # Y-position of the center of mass (Yna) measured from the base
    y_cg = ( (t_base + h_web + 0.5*t_flange)*area_flange + (t_base + 0.5*h_web)*area_web + 0.5*t_base*area_base ) / area
    return area, y_cg


def _compute_applied_axial(R_od, t_wall, m_stack, section_mass):
    """Compute axial stress for spar from z-axis loading

    INPUTS:
    ----------
    params       : dictionary of input parameters
    section_mass : float (scalar/vector),  mass of each spar section as axial loading increases with spar depth

    OUTPUTS:
    -------
    stress   : float (scalar/vector),  axial stress
    """

    R       = R_od - 0.5*t_wall
    # Add in weight of sections above it
    axial_load    = m_stack + np.r_[0.0, np.cumsum(section_mass[:-1])]
    # Divide by shell cross sectional area to get stress
    return (gravity * axial_load / (2.0 * np.pi * R * t_wall))


def _compute_applied_hoop(pressure, R_od, t_wall):
    """Compute hoop stress WITHOUT accounting for stiffener rings

    INPUTS:
    ----------
    pressure : float (scalar/vector),  radial (hydrostatic) pressure
    R_od     : float (scalar/vector),  radius to outer wall of shell
    t_wall   : float (scalar/vector),  shell wall thickness

    OUTPUTS:
    -------
    stress   : float (scalar/vector),  hoop stress with no stiffeners
    """
    return (pressure * R_od / t_wall)

    
def _compute_stiffener_factors(pressure, axial_stress, R_od, t_wall, h_web, t_web, w_flange, t_flange, L_stiffener, E, nu):
    """Compute modifiers to stress due to presence of stiffener rings.

    INPUTS:
    ----------
    params       : dictionary of input parameters
    pressure     : float (scalar/vector),  radial (hydrostatic) pressure
    axial_stress : float (scalar/vector),  axial loading (z-axis) stress

    OUTPUTS:
    -------
    stiffener_factor_KthL : float (scalar/vector),  Stress modifier from stiffeners for local buckling from axial loads
    stiffener_factor_KthG : float (scalar/vector),  Stress modifier from stiffeners for general buckling from external pressure
    """

    # Geometry computations
    R_flange = R_od - h_web # Should have "- t_wall", but not in appendix B
    area_stiff, y_cg, Ixx, Iyy = _TBeamProperties(h_web, t_web, w_flange, t_flange)
    t_stiff  = area_stiff / h_web # effective thickness(width) of stiffener section

    # Compute hoop stress modifiers accounting for stiffener rings
    # This has to be done at midpoint between stiffeners and at stiffener location itself
    # Compute beta (just a local term used here)
    D    = E * t_wall**3 / (12.0 * (1 - nu*nu))
    beta = (0.25 * E * t_wall / R_od**2 / D)**0.25
    # Compute psi-factor (just a local term used here)
    u     = np.minimum(0.5 * beta * L_stiffener, 30.0)
    psi_k = 2.0 * (np.sin(u)*np.cosh(u) + np.cos(u)*np.sinh(u)) / (np.sinh(2*u) + np.sin(2*u)) 

    # Compute a couple of other local terms
    u   = np.minimum(beta * L_stiffener, 30.0)
    k_t = 8 * beta**3 * D * (np.cosh(u) - np.cos(u)) / (np.sinh(u) + np.sin(u))
    k_d = E * t_stiff * (R_od**2 - R_flange**2) / R_od / ((1+nu)*R_od**2 + (1-nu)*R_flange**2)

    # Pressure from axial load
    pressure_sigma = pressure - nu*axial_stress*t_wall/R_od

    # Compute the correction to hoop stress due to the presesnce of ring stiffeners
    stiffener_factor_KthL = 1 - psi_k * (pressure_sigma / pressure) * (k_d / (k_d + k_t))
    stiffener_factor_KthG = 1 -         (pressure_sigma / pressure) * (k_d / (k_d + k_t))
    return stiffener_factor_KthL, stiffener_factor_KthG


def _compute_elastic_stress_limits(R_od, t_wall, h_section, h_web, t_web, w_flange, t_flange,
                                   L_stiffener, E, nu, KthG, loading='hydrostatic'):
    """Compute modifiers to stress due to presence of stiffener rings.

    INPUTS:
    ----------
    params  : dictionary of input parameters
    KthG    : float (scalar/vector),  Stress modifier from stiffeners for general buckling from external pressure
    loading : string (hydrostatic/radial), Parameter that determines a coefficient- is only included for unit testing 
              consistency with API 2U Appdx B and should not be used in practice

    OUTPUTS:
    -------
    elastic_axial_local_FxeL    : float (scalar/vector),  Elastic stress limit for local buckling from axial loads
    elastic_extern_local_FreL   : float (scalar/vector),  Elastic stress limit for local buckling from external pressure loads
    elastic_axial_general_FxeG  : float (scalar/vector),  Elastic stress limit for general instability from axial loads
    elastic_extern_general_FreG : float (scalar/vector),  Elastic stress limit for general instability from external pressure loads
    """

    # Geometry computations
    nsections = R_od.size
    area_stiff, y_cg, Ixx, Iyy = _TBeamProperties(h_web, t_web, w_flange, t_flange)
    area_stiff_bar = area_stiff / L_stiffener / t_wall
    R  = R_od - 0.5*t_wall

    # 1. Local shell mode buckling from axial loads
    # Compute a few parameters that define the curvature of the geometry
    m_x  = L_stiffener / np.sqrt(R * t_wall)
    z_x  = m_x**2 * np.sqrt(1 - nu**2)
    z_m  = 12.0 * z_x**2 / np.pi**4
    # Imperfection factor- empirical fit that converts theory to reality
    a_xL = 9.0 * (300.0 + (2*R/t_wall))**(-0.4)
    # Calculate buckling coefficient
    C_xL = np.sqrt( 1 + 150.0 * a_xL**2 * m_x**4 / (2*R/t_wall) )
    # Calculate elastic and inelastic final limits
    elastic_axial_local_FxeL   = C_xL * np.pi**2 * E * (t_wall/L_stiffener)**2 / 12.0 / (1-nu**2)

    # 2. Local shell mode buckling from external (pressure) loads
    # Imperfection factor- empirical fit that converts theory to reality
    a_thL = np.ones(m_x.shape)
    a_thL[m_x > 5.0] = 0.8
    # Find the buckling mode- closest integer that is root of solved equation
    n   = np.zeros((nsections,))
    maxn = 50
    for k in range(nsections):
        c = L_stiffener[k] / np.pi / R[k]
        myfun = lambda x:((c*x)**2*(1 + (c*x)**2)**4/(2 + 3*(c*x)**2) - z_m[k])
        try:
            n[k] = brentq(myfun, 0, maxn)
        except:
            n[k] = maxn
    # Calculate beta (local term)
    beta  = np.round(n) * L_stiffener / np.pi / R
    # Calculate buckling coefficient
    C_thL = a_thL * ( (1+beta**2)**2/(0.5+beta**2) + 0.112*m_x**4/(1+beta**2)**2/(0.5+beta**2) )
    # Calculate elastic and inelastic final limits
    elastic_extern_local_FreL   = C_thL * np.pi**2 * E * (t_wall/L_stiffener)**2 / 12.0 / (1-nu**2)

    # 3. General instability buckling from axial loads
    # Compute imperfection factor
    a_x = 0.85 / (1 + 0.0025*2*R/t_wall)
    a_xG = a_x
    a_xG[area_stiff_bar>=0.2] = 0.72
    ind = np.logical_and(area_stiff_bar<0.06, area_stiff_bar<0.2)
    a_xG[ind] = (3.6 - 5.0*a_x[ind])*area_stiff_bar[ind]
    # Calculate elastic and inelastic final limits
    elastic_axial_general_FxeG   = 0.605 * a_xG * E * t_wall/R * np.sqrt(1 + area_stiff_bar)

    # 4. General instability buckling from external loads
    # Distance from shell centerline to stiffener cg
    z_r = -(y_cg + 0.5*t_wall)
    # Imperfection factor
    a_thG = 0.8
    # Effective shell width if the outer shell and the T-ring stiffener were to be combined to make an uneven I-beam
    L_shell_effective = 1.1*np.sqrt(2.0*R*t_wall) + t_web
    L_shell_effective[m_x <= 1.56] = L_stiffener[m_x <= 1.56]
    # Get properties of this effective uneven I-beam
    _, yna_eff = _IBeamProperties(h_web, t_web, w_flange, t_flange, L_shell_effective, t_wall)
    Rc = R_od - yna_eff
    # Compute effective shell moment of inertia based on Ir - I of stiffener
    Ier = Ixx + area_stiff*z_r**2*L_shell_effective*t_wall/(area_stiff+L_shell_effective*t_wall) + L_shell_effective*t_wall**3/12.0
    # Lambda- a local constant
    lambda_G = np.pi * R / h_section
    # Coefficient factor listed as 'k' in peG equation
    coeff = 0.5 if loading in ['hydro','h','hydrostatic','static'] else 0.0    
    # Compute pressure leading to elastic failure
    n = np.zeros(R_od.shape)
    pressure_failure_peG = np.zeros(R_od.shape)
    for k in range(nsections):
        peG = lambda x: ( E*lambda_G[k]**4*t_wall[k]/R[k]/(x**2+0.0*lambda_G[k]**2-1)/(x**2 + lambda_G[k]**2)**2 +
                          E*Ier[k]*(x**2-1)/L_stiffener[k]/Rc[k]**2/R_od[k] )
        minout = minimize_scalar(peG, bounds=(2.0, 15.0), method='bounded')
        n[k] = minout.x
        pressure_failure_peG[k] = peG(n[k])
    # Calculate elastic and inelastic final limits
    elastic_extern_general_FreG   = a_thG * pressure_failure_peG * R_od * KthG / t_wall

    return elastic_axial_local_FxeL, elastic_extern_local_FreL, elastic_axial_general_FxeG, elastic_extern_general_FreG




def _plasticityRF(Felastic, yield_stress):
    """Computes plasticity reduction factor for elastic stresses near the yield stress to obtain an inelastic stress
    This is defined in Section 5 of API Bulletin 2U

    INPUTS:
    ----------
    Felastic     : float (scalar/vector),  elastic stress
    yield_stress : float (scalar/vector),  yield stress

    OUTPUTS:
    -------
    Finelastic   : float (scalar/vector),  modified (in)elastic stress
    """
    Fratio = np.array(yield_stress / Felastic)
    eta    = Fratio * (1.0 + 3.75*Fratio**2)**(-0.25)
    Finelastic = np.array(Felastic)
    Finelastic[Felastic > 0.5*yield_stress] *= eta[Felastic > 0.5*yield_stress]
    return Finelastic


def _safety_factor(Ficj, yield_stress):
    """Use the inelastic limits and yield stress to compute required safety factors
    This is defined in Section 9 of API Bulletin 2U

    INPUTS:
    ----------
    Ficj          : float (scalar/vector),  inelastic stress
    yield_stress  : float (scalar/vector),  yield stress

    OUTPUTS:
    -------
    safety_factor : float (scalar/vector),  margin applied to inelastic stress limits
    """
    # Partial safety factor, psi
    psi = np.array(1.4 - 0.4 * Ficj / yield_stress)
    psi[Ficj <= 0.5*yield_stress] = 1.2
    psi[Ficj >= yield_stress] = 1.0
    # Final safety factor is 25% higher to give a margin
    return 1.25*psi


def shellBuckling_withStiffeners(P, sigma_ax, R_od, t_wall, h_section, h_web, t_web, w_flange, t_flange,
                                 L_stiffener, E, nu, sigma_y, loading='hydro'):

    # APPLIED STRESSES (Section 11 of API Bulletin 2U)
    stiffener_factor_KthL, stiffener_factor_KthG = _compute_stiffener_factors(P, sigma_ax, R_od, t_wall, h_web, t_web,
                                                                              w_flange, t_flange, L_stiffener, E, nu)
    hoop_stress_nostiff = _compute_applied_hoop(P, R_od, t_wall)
    hoop_stress_between = hoop_stress_nostiff * stiffener_factor_KthL
    hoop_stress_atring  = hoop_stress_nostiff * stiffener_factor_KthG
    
    # BUCKLING FAILURE STRESSES (Section 4 of API Bulletin 2U)
    (elastic_axial_local_FxeL, elastic_extern_local_FreL,
     elastic_axial_general_FxeG, elastic_extern_general_FreG) = _compute_elastic_stress_limits(R_od, t_wall, h_section, h_web,
                                                                                               t_web, w_flange, t_flange,
                                                                                               L_stiffener, E, nu, stiffener_factor_KthG,
                                                                                               loading=loading)
    inelastic_axial_local_FxcL    = _plasticityRF(elastic_axial_local_FxeL   , sigma_y)
    inelastic_axial_general_FxcG  = _plasticityRF(elastic_axial_general_FxeG , sigma_y)
    inelastic_extern_local_FrcL   = _plasticityRF(elastic_extern_local_FreL  , sigma_y)
    inelastic_extern_general_FrcG = _plasticityRF(elastic_extern_general_FreG, sigma_y)

    # COMBINE AXIAL AND HOOP (EXTERNAL PRESSURE) LOADS TO FIND DESIGN LIMITS
    # (Section 6 of API Bulletin 2U)
    load_per_length_Nph = sigma_ax            * t_wall
    load_per_length_Nth = hoop_stress_nostiff * t_wall
    load_ratio_k        = load_per_length_Nph / load_per_length_Nth
    def solveFthFph(Fxci, Frci, Kth):
        Fphci = np.zeros(Fxci.shape)
        Fthci = np.zeros(Fxci.shape)
        Kph   = 1.0
        c1    = (Fxci + Frci) / sigma_y - 1.0
        c2    = load_ratio_k * Kph / Kth
        for k in range(Fxci.size):
            try:
                Fthci[k] = brentq(lambda x: (c2[k]*x/Fxci[k])**2 - c1[k]*(c2[k]*x/Fxci[k])*(x/Frci[k]) + (x/Frci[k])**2 - 1.0, 0, Fxci[k]+Frci[k], maxiter=20)
            except:
                Fthci[k] = Fxci[k] + Frci[k]
            Fphci[k] = c2[k] * Fthci[k]
        return Fphci, Fthci

    inelastic_local_FphcL, inelastic_local_FthcL = solveFthFph(inelastic_axial_local_FxcL, inelastic_extern_local_FrcL, stiffener_factor_KthL)
    inelastic_general_FphcG, inelastic_general_FthcG = solveFthFph(inelastic_axial_general_FxcG, inelastic_extern_general_FrcG, stiffener_factor_KthG)

    # Use the inelastic limits and yield stress to compute required safety factors
    # and adjust the limits accordingly
    axial_limit_local_FaL     = inelastic_local_FphcL   / _safety_factor(inelastic_local_FphcL  , sigma_y)
    extern_limit_local_FthL   = inelastic_local_FthcL   / _safety_factor(inelastic_local_FthcL  , sigma_y)
    axial_limit_general_FaG   = inelastic_general_FphcG / _safety_factor(inelastic_general_FphcG, sigma_y)
    extern_limit_general_FthG = inelastic_general_FthcG / _safety_factor(inelastic_general_FthcG, sigma_y)

    # Compare limits to applied stresses and use this ratio as a design constraint
    # (Section 9 "Allowable Stresses" of API Bulletin 2U)
    # These values must be <= 1.0
    axial_local_api      = sigma_ax / axial_limit_local_FaL
    axial_general_api    = sigma_ax / axial_limit_general_FaG
    external_local_api   = hoop_stress_between / extern_limit_local_FthL
    external_general_api = hoop_stress_between / extern_limit_general_FthG

    # Compute unification ratios without safety factors in case we want to apply our own later
    axial_local_raw      = sigma_ax / inelastic_local_FphcL
    axial_general_raw    = sigma_ax / inelastic_general_FphcG
    external_local_raw   = hoop_stress_between / inelastic_local_FthcL
    external_general_raw = hoop_stress_between / inelastic_general_FthcG
    
    return (axial_local_api, axial_general_api, external_local_api, external_general_api,
            axial_local_raw, axial_general_raw, external_local_raw, external_general_raw)



"""DFIG.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved."""

from openmdao.api import Group, Problem, ExplicitComponent,ExecComp,IndepVarComp,ScipyOptimizeDriver
import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan
from dfigSE import DFIG

class DFIG_OM(ExplicitComponent):
    """ Estimates overall mass, dimensions and Efficiency of DFIG generator. """
    
    def setup(self):
        # DFIG design inputs
        #self.add_input('r_s', val=0.0, units='m', desc='airgap radius r_s')
        self.add_input('rad_ag', val=0.0, units='m', desc='airgap radius')
        self.add_input('len_s',  val=0.0, units='m', desc='Stator core length l_s')
        self.add_input('h_s',    val=0.0, units='m', desc='Stator slot height')
        self.add_input('h_r',    val=0.0, units='m', desc='Rotor slot height')
        
        self.add_input('machine_rating', val=0.0, units='W',   desc='Machine rating')
        self.add_input('n_nom',          val=0.0, units='rpm', desc='rated speed')
        
        self.add_input('B_symax', val=0.0,            desc='Peak Stator Yoke flux density B_symax')
        self.add_input('S_Nmax',  val=0.0, units='T', desc='Max rated Slip ')
        self.add_input('I_0',     val=0.0, units='A', desc='no-load excitation current')
        
        self.add_input('shaft_cm',           val= np.zeros(3), units='m', desc='Main Shaft CM')
        self.add_input('shaft_length',       val=0.0,          units='m', desc='main shaft length')
        self.add_input('Gearbox_efficiency', val=0.0,                     desc='Gearbox efficiency')
        
        # Material properties
        self.add_input('rho_Fe',     val=0.0, units='kg*m**-3', desc='Magnetic Steel density ')
        self.add_input('rho_Copper', val=0.0, units='kg*m**-3', desc='Copper density ')
        
        # DFIG generator design output
        # Magnetic loading
        self.add_output('B_g',     val=0.0, desc='Peak air gap flux density B_g')
        self.add_output('B_g1',    val=0.0, desc='air gap flux density fundamental ')
        self.add_output('B_rymax', val=0.0, desc='maximum flux density in rotor yoke')
        self.add_output('B_tsmax', val=0.0, desc='maximum tooth flux density in stator')
        self.add_output('B_trmax', val=0.0, desc='maximum tooth flux density in rotor')
        
        #Stator design
        self.add_output('N_s',                val=0.0, desc='Stator turns')
        self.add_output('S',                  val=0.0, desc='Stator slots')
        self.add_output('h_ys',               val=0.0, desc='Stator Yoke height')
        self.add_output('b_s',                val=0.0, desc='stator slot width')
        self.add_output('b_t',                val=0.0, desc='stator tooth width')
        self.add_output('A_Cuscalc',          val=0.0, desc='Stator Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio1', val=0.0, desc='Stator slot aspect ratio')
        self.add_output('D_ratio',            val=0.0, desc='Stator diameter ratio')
        self.add_output('q1',                 val=0.0, desc='Slots per pole per phase')
        
        #Rotor design
        self.add_output('N_r',                val=0.0, desc='Rotor turns')
        self.add_output('Q_r',                val=0.0, desc='Rotor slots')
        self.add_output('h_yr',               val=0.0, desc='rotor yoke height')
        self.add_output('b_r',                val=0.0, desc='rotor slot width')
        self.add_output('b_tr',               val=0.0, desc='rotor tooth width')
        self.add_output('A_Curcalc',          val=0.0, desc='Rotor Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio2', val=0.0, desc='Rotor slot aspect ratio')
        self.add_output('b_trmin',            val=0.0, desc='minimum tooth width')
        self.add_output('tau_p',              val=0.0, desc='Pole pitch')
        self.add_output('p',                  val=0.0, desc='No of pole pairs')
        
        # Electrical performance
        self.add_output('E_p',  val=0.0, desc='Stator phase voltage')
        self.add_output('f',    val=0.0, desc='Generator output frequency')
        self.add_output('I_s',  val=0.0, desc='Generator output phase current')
        self.add_output('A_1' , val=0.0, desc='Specific current loading')
        self.add_output('J_s',  val=0.0, desc='Stator winding Current density')
        self.add_output('J_r',  val=0.0, desc='Rotor winding Current density')
        self.add_output('R_s',  val=0.0, desc='Stator resistance')
        self.add_output('R_R',  val=0.0, desc='Rotor resistance')
        self.add_output('L_r',  val=0.0, desc='Rotor inductance')
        self.add_output('L_s',  val=0.0, desc='Stator synchronising inductance')
        self.add_output('L_sm', val=0.0, desc='mutual inductance')
        
        # Objective functions
        self.add_output('Mass',    val=0.0, desc='Actual mass')
        self.add_output('K_rad',   val=0.0, desc='Stack length ratio')
        self.add_output('Losses',  val=0.0, desc='Total loss')
        self.add_output('gen_eff', val=0.0, desc='Generator efficiency')
        
        # Mass Outputs
        self.add_output('Copper',          val=0.0, units='kg', desc='Copper Mass')
        self.add_output('Iron',            val=0.0, units='kg', desc='Electrical Steel Mass')
        self.add_output('Structural_mass', val=0.0, units='kg', desc='Structural Mass')
        
        # Structural performance
        self.add_output('TC1', val=0.0, desc='Torque constraint-stator')
        self.add_output('TC2', val=0.0, desc='Torque constraint-rotor')
        
        # Other parameters
        self.add_output('Current_ratio', val=0.0,         desc='Rotor current ratio')
        self.add_output('Overall_eff',   val=0.0,         desc='Overall drivetrain efficiency')
        self.add_output('I',             val=np.zeros(3), desc='Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('cm',            val=np.zeros(3), desc='COM [x,y,z]')
        
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack inputs
        rad_ag               = inputs['rad_ag']
        len_s                = inputs['len_s']
        h_s                  = inputs['h_s']
        h_r                  = inputs['h_r']
        I_0                  = inputs['I_0']
        
        machine_rating       = inputs['machine_rating']
        n_nom                = inputs['n_nom']
        Gearbox_efficiency   = inputs['Gearbox_efficiency']
        
        rho_Fe               = inputs['rho_Fe']
        rho_Copper           = inputs['rho_Copper']
        
        B_symax              = inputs['B_symax']
        S_Nmax               = inputs['S_Nmax']
        
        shaft_cm             = inputs['shaft_cm']
        shaft_length         = inputs['shaft_length']
        
        dfig = DFIG()
        
        (outputs['B_g'], outputs['B_g1'], outputs['B_rymax'], outputs['B_tsmax'], outputs['B_trmax'], outputs['q1'], 
        outputs['N_s'], outputs['S'], outputs['h_ys'], outputs['b_s'], outputs['b_t'], outputs['D_ratio'], 
        outputs['A_Cuscalc'], outputs['Slot_aspect_ratio1'], outputs['h_yr'], outputs['tau_p'], outputs['p'], 
        outputs['Q_r'], outputs['N_r'], outputs['b_r'], outputs['b_trmin'], outputs['b_tr'], outputs['A_Curcalc'],
        outputs['Slot_aspect_ratio2'], outputs['E_p'], outputs['f'], outputs['I_s'], outputs['A_1'], outputs['J_s'], 
        outputs['J_r'], outputs['R_s'], outputs['R_R'], outputs['L_r'], outputs['L_s'], outputs['L_sm'], 
        outputs['Mass'], outputs['K_rad'], outputs['Losses'], outputs['gen_eff'], outputs['Copper'], 
        outputs['Iron'], outputs['Structural_mass'], outputs['TC1'], outputs['TC2'], outputs['Current_ratio'], 
        outputs['Overall_eff'], outputs['cm'], outputs['I']) \
            = dfig.compute(rad_ag, len_s, h_s, h_r, I_0, machine_rating, n_nom, Gearbox_efficiency, 
            rho_Fe, rho_Copper, B_symax, S_Nmax, shaft_cm, shaft_length)
            



"""dfigSE.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved.

The pure-python part of a doubly-fed induction generator module.
The OpenMDAO part is in dfigOM.py
"""

import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan
import sys

class DFIG(object):
    """ Estimates overall mass, dimensions and Efficiency of DFIG generator. """
    
    def __init__(self):

        super(DFIG, self).__init__()
        self.debug = False

    #------------------------
        
    def compute(self, rad_ag, len_s, h_s, h_r, I_0, machine_rating, n_nom, Gearbox_efficiency, 
        rho_Fe, rho_Copper, B_symax, S_Nmax, shaft_cm, shaft_length,
                debug=False):
      
        
        #Assign values to universal constants
        g1          = 9.81             # m/s^2 acceleration due to gravity
        sigma       = 21.5e3           # shear stress in psi (what material? Al, brass, Cu?) ~148e6 Pa
        sys.stderr.write('\ndfigSE: Check units on sigma!\n\n')
        mu_0        = pi * 4e-7        # permeability of free space in m * kg / (s**2 * A**2)
        cofi        = 0.9              # power factor
        h_w         = 0.005            # wedge height
        m           = 3                # Number of phases
        resist_Cu   = 1.8e-8 * 1.4     # copper resisitivity
        h_sy0       = 0
        
        #Assign values to design constants
        b_so = 0.004                    # Stator slot opening width
        b_ro = 0.004                    # Rotor  slot opening width
        q1 = 5                          # Stator slots per pole per phase
        q2 = q1 - 1                     # Rotor  slots per pole per phase
        b_s_tau_s = 0.45                # Stator slot-width / slot-pitch ratio
        b_r_tau_r = 0.45                # Rotor  slot-width / slot-pitch ratio
        y_tau_p  = 12. / 15             # Stator coil span to pole pitch    
        y_tau_pr = 10. / 12             # Rotor  coil span to pole pitch
        
        p = 3                           # pole pairs
        freq = 60                       # grid frequency in Hz
        k_fillr = 0.55                  # Rotor Slot fill factor
        k_fills = 0.65                  # Stator Slot fill factor (will be set later)
        P_Fe0h = 4                      # specific hysteresis losses W / kg @ 1.5 T 
        P_Fe0e = 1                      # specific eddy losses W / kg @ 1.5 T 
        
        K_rs     = 1 / ( -1 * S_Nmax)              # Winding turns ratio between rotor and Stator        
        I_SN   = machine_rating / (sqrt(3) * 3000) # Rated current
        I_SN_r = I_SN / K_rs                       # Stator rated current reduced to rotor
        
        # Calculating winding factor for stator and rotor    
        
        k_y1 = sin(pi/2 * y_tau_p)                              # winding chording factor
        k_q1 = sin(pi/6) / (q1 * sin(pi/(6 * q1)))              # winding zone factor
        k_y2 = sin(pi/2 * y_tau_pr)                             # winding chording factor
        k_q2 = sin(pi/6) / (q2 * sin(pi/(6 * q2)))              # winding zone factor
        k_wd1 = k_y1 * k_q1                                     # Stator winding factor
        k_wd2 = k_q2 * k_y2                                     # Rotor winding factor
        
        ag_dia = 2 * rad_ag                                     # air gap diameter
        ag_len = (0.1 + 0.012 * machine_rating**(1./3)) * 0.001 # air gap length in m
        K_rad = len_s / ag_dia                                  # Aspect ratio
        rad_r = rad_ag - ag_len                                 # rotor radius  (was r_r)
        tau_p = pi * ag_dia / (2 * p)                           # pole pitch
        
        S = 2 * p * q1 * m                                      # Stator slots
        N_slots_pp = S / (m * p * 2)                            # Number of stator slots per pole per phase
        n  = S / 2 * p / q1                                     # no of slots per pole per phase
        tau_s = tau_p / (m * q1)                                # Stator slot pitch
        b_s = b_s_tau_s * tau_s                                 # Stator slot width
        b_t = tau_s - b_s                                       # Stator tooth width
        
        Q_r = 2 * p * m * q2                                    # Rotor slots
        tau_r = pi * (ag_dia - 2 * ag_len) / Q_r                # Rotor slot pitch
        b_r = b_r_tau_r * tau_r                                 # Rotor slot width
        b_tr = tau_r - b_r                                      # Rotor tooth width
        
        # Calculating equivalent slot openings
        
        mu_rs = 0.005
        mu_rr = 0.005
        W_s   = (b_s / mu_rs) * 1e-3  # Stator, in m
        W_r   = (b_r / mu_rr) * 1e-3  # Rotor,  in m
        
        Slot_aspect_ratio1 = h_s / b_s
        Slot_aspect_ratio2 = h_r / b_r
        
        # Calculating Carter factor for stator,rotor and effective air gap length
        
        gamma_s = (2 * W_s / ag_len)**2 / (5 + 2 * W_s / ag_len)
        K_Cs    = tau_s / (tau_s - ag_len * gamma_s * 0.5)  #page 3 - 13
        gamma_r = (2 * W_r / ag_len)**2 / (5 + 2 * W_r / ag_len)
        K_Cr    = tau_r / (tau_r - ag_len * gamma_r * 0.5)  #page 3 - 13
        K_C     = K_Cs * K_Cr
        g_eff   = K_C * ag_len
        
        om_m = 2 * pi * n_nom / 60                      # mechanical frequency
        om_e = p * om_m                                 # electrical frequency
        f = n_nom * p / 60                              # generator output freq
        K_s = 0.3                                       # saturation factor for Iron
        n_c1 = 2                                        # number of conductors per coil
        a1 = 2                                          # number of parallel paths
        N_s = np.round(2 * p * N_slots_pp * n_c1 / a1)  # Stator winding turns per phase        
        N_r = np.round(N_s * k_wd1 * K_rs / k_wd2)      # Rotor winding turns per phase
        n_c2 = N_r / (Q_r / m)                          # rotor turns per coil
        
        # Calculating peak flux densities and back iron thickness
        
        B_g1 = mu_0 * 3 * N_r * I_0 * sqrt(2) * k_y2 * k_q2 / (pi * p * g_eff * (1 + K_s))
        B_g = B_g1 * K_C
        h_ys = B_g * tau_p / (B_symax * pi)
        B_rymax = B_symax
        h_yr = h_ys
        B_tsmax = B_g * tau_s / b_t 
        
        d_se = ag_dia + 2 * (h_ys + h_s + h_w)           # stator outer diameter
        D_ratio = d_se / ag_dia                          # Diameter ratio
        
        # Stator slot fill factor
        if ag_dia > 2:
            k_fills = 0.65
        else:
            k_fills = 0.4
            
        # Stator winding calculation
        
        # End connection length for stator winding coils
        
        l_fs = 2 * (0.015 + y_tau_p * tau_p / (2 * cos(radians(40)))) + pi * h_s   # added radians() 2019 09 11
        
        l_Cus = 2 * N_s * (l_fs + len_s) / a1             # Length of Stator winding 
        
        # Conductor cross-section
        A_s       = b_s *        (h_s - h_w)
        A_scalc   = b_s * 1000 * (h_s - h_w) * 1000
        A_Cus     = A_s     * q1 * p * k_fills / N_s
        A_Cuscalc = A_scalc * q1 * p * k_fills / N_s
        
        # Stator winding resistance
        
        R_s = l_Cus * resist_Cu / A_Cus
        tau_r_min = pi * (ag_dia - 2 * (ag_len + h_r)) / Q_r
        
        # Peak magnetic loading on the rotor tooth
        
        b_trmin = tau_r_min - b_r_tau_r * tau_r_min
        B_trmax = B_g * tau_r / b_trmin
        
        # Calculating leakage inductance in  stator
        
        K_01 = 1 - 0.033 * (W_s**2 / ag_len / tau_s)
        sigma_ds = 0.0042
        
        L_ssigmas  = (2 * mu_0 * len_s * n_c1**2 * S / m / a1**2) * ((h_s - h_w) / (3 * b_s) + h_w / b_so)                # slot leakage inductance
        L_ssigmaew = (2 * mu_0 * len_s * n_c1**2 * S / m / a1**2) * 0.34 * q1 * (l_fs - 0.64 * tau_p * y_tau_p) / len_s   # end winding leakage inductance
        L_ssigmag  = (2 * mu_0 * len_s * n_c1**2 * S / m / a1**2) * (0.9 * tau_s * q1 * k_wd1 * K_01 * sigma_ds / g_eff)  # tooth tip leakage inductance
        L_ssigma   = (L_ssigmas + L_ssigmaew + L_ssigmag)                                                                 # stator leakage inductance
        L_sm       = 6 * mu_0 * len_s * tau_p * (k_wd1 * N_s)**2 / (pi**2 * (p) * g_eff * (1 + K_s))                      
        L_s        = (L_ssigmas + L_ssigmaew + L_ssigmag)                                                                 # stator  inductance
        
        # Calculating leakage inductance in  rotor
        
        K_02 = 1 - 0.033 * (W_r**2 / ag_len / tau_r)
        sigma_dr = 0.0062
        
        l_fr = (0.015 + y_tau_pr * tau_r / 2 / cos(radians(40))) + pi * h_r                                       # Rotor end connection length
        L_rsl = (mu_0 * len_s * (2 * n_c2)**2 * Q_r / m) * ((h_r - h_w) / (3 * b_r) + h_w / b_ro)                 # slot leakage inductance
        L_rel = (mu_0 * len_s * (2 * n_c2)**2 * Q_r / m) * 0.34 * q2 * (l_fr - 0.64 * tau_r * y_tau_pr) / len_s   # end winding leakage inductance
        L_rtl = (mu_0 * len_s * (2 * n_c2)**2 * Q_r / m) * (0.9 * tau_s * q2 * k_wd2 * K_02 * sigma_dr / g_eff)   # tooth tip leakage inductance
        L_r = (L_rsl + L_rtl + L_rel) / K_rs**2                                                                   # rotor leakage inductance
        sigma1 = 1 - (L_sm**2 / L_s / L_r)
        
        # Rotor Field winding
        
        # conductor cross-section
        diff = h_r - h_w
        A_Cur = k_fillr * p*q2 * b_r * diff / N_r
        A_Curcalc = A_Cur * 1e6
        
        L_cur = 2 * N_r * (l_fr + len_s)      # rotor winding length
        Resist_r = resist_Cu * L_cur / A_Cur  # Rotor resistance
        R_R = Resist_r / K_rs**2              # Equivalent rotor resistance reduced to stator
        
        om_s = n_nom * 2*pi / 60                     # synchronous speed in rad / s
        P_e = machine_rating / (1 - S_Nmax)          # Air gap power
        
        # Calculating No-load voltage
        E_p      = om_s * N_s * k_wd1 * rad_ag * len_s * B_g1 * sqrt(2)
        I_r      = P_e / m / E_p                        # rotor active current        
        I_sm     = E_p / (2 * pi * freq * (L_s + L_sm)) # stator reactive current
        I_s      = sqrt(I_r**2 + I_sm**2)               # Stator current
        I_srated = machine_rating / 3 / K_rs / E_p      # Rated current
        
        # Calculating winding current densities and specific current loading
        
        J_s = I_s / A_Cuscalc
        J_r = I_r / A_Curcalc
        A_1 = 2 * m*N_s * I_s / (pi * 2 * rad_ag)
        Current_ratio = I_0 / I_srated           # Ratio of magnetization current to rated current
        
        # Calculating masses of the electromagnetically active materials
        
        V_Cuss = m * l_Cus * A_Cus
        V_Cusr = m * L_cur * A_Cur
        V_Fest = len_s * pi * ((rad_ag + h_s)**2 - rad_ag**2) - (2 * m*q1 * p*b_s * h_s * len_s)
        V_Fesy = len_s * pi * ((rad_ag + h_s + h_ys)**2    - (rad_ag + h_s)**2)
        V_Fert = len_s * pi * (rad_r**2 - (rad_r - h_r)**2) - 2 * m*q2 * p*b_r * h_r * len_s
        V_Fery = len_s * pi * ((rad_r - h_r)**2           - (rad_r - h_r - h_yr)**2)
        Copper = (V_Cuss + V_Cusr) * rho_Copper
        M_Fest = V_Fest * rho_Fe
        M_Fesy = V_Fesy * rho_Fe
        M_Fert = V_Fert * rho_Fe
        M_Fery = V_Fery * rho_Fe
        Iron = M_Fest + M_Fesy + M_Fert + M_Fery
        M_gen = (Copper) + (Iron)
        
        #K_gen = Cu * C_Cu + (Iron) * C_Fe #%M_pm * K_pm
        
        L_tot = len_s
        Structural_mass = 0.0002 * M_gen**2 + 0.6457 * M_gen + 645.24
        Mass = M_gen + Structural_mass
        
        # Calculating Losses and efficiency 
        # 1. Copper losses
        
        K_R = 1.2 # skin effect correction coefficient
        P_Cuss = m * I_s**2 * R_s * K_R            # Copper loss - stator
        P_Cusr = m * I_r**2 * R_R                  # Copper loss - rotor
        P_Cusnom = P_Cuss + P_Cusr                 # Copper loss - total
        
        # Iron Losses ( from Hysteresis and eddy currents)      
        P_Hyys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))                    # Hysteresis losses in stator yoke
        P_Ftys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)               # Eddy losses in stator yoke
        P_Hyd  = M_Fest * (B_tsmax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))                    # Hysteresis losses in stator teeth
        P_Ftd  = M_Fest * (B_tsmax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)               # Eddy losses in stator teeth
        P_Hyyr = M_Fery * (B_rymax / 1.5)**2 * (P_Fe0h * abs(S_Nmax) * om_e / (2 * pi * 60))      # Hysteresis losses in rotor yoke
        P_Ftyr = M_Fery * (B_rymax / 1.5)**2 * (P_Fe0e * (abs(S_Nmax) * om_e / (2 * pi * 60))**2) # Eddy losses in rotor yoke
        P_Hydr = M_Fert * (B_trmax / 1.5)**2 * (P_Fe0h * abs(S_Nmax) * om_e / (2 * pi * 60))      # Hysteresis losses in rotor teeth
        P_Ftdr = M_Fert * (B_trmax / 1.5)**2 * (P_Fe0e * (abs(S_Nmax) * om_e / (2 * pi * 60))**2) # Eddy losses in rotor teeth
        P_add = 0.5 * machine_rating / 100                                                        # additional losses
        P_Fesnom = P_Hyys + P_Ftys + P_Hyd + P_Ftd + P_Hyyr + P_Ftyr + P_Hydr + P_Ftdr            # Total iron loss
        delta_v = 1                                                                               # allowable brush voltage drop
        p_b = 3 * delta_v * I_r                                                                   # Brush loss
        
        Losses = P_Cusnom + P_Fesnom + p_b + P_add
        gen_eff = (P_e - Losses) * 100 / P_e
        Overall_eff = gen_eff * Gearbox_efficiency
        
        # Calculating stator winding current density
        J_s = I_s / A_Cuscalc
        
        # Calculating  electromagnetic torque
        T_e = p *(machine_rating * 1.01) / (2 * pi * freq * (1 - S_Nmax))
        
        # Calculating for tangential stress constraints
        
        TC1 = T_e / (2 * pi * sigma)
        TC2 = rad_ag**2 * len_s
        
        # Calculating mass moments of inertia and center of mass
        I = np.zeros(3)
        r_out = d_se * 0.5
        I[0]   = 0.50 * Mass * r_out**2
        I[1]   = 0.25 * Mass * r_out**2 + Mass * len_s**2 / 12
        I[2]   = I[1]
        cm = np.zeros(3)
        cm[0]  = shaft_cm[0] + shaft_length/2. + len_s / 2
        cm[1]  = shaft_cm[1]
        cm[2]  = shaft_cm[2]

        return  B_g, B_g1, B_rymax, B_tsmax, B_trmax, q1,  N_s, S, h_ys, b_s, b_t, D_ratio, \
            A_Cuscalc, Slot_aspect_ratio1, h_yr, tau_p, p,  Q_r, N_r, b_r, b_trmin, b_tr, A_Curcalc, \
            Slot_aspect_ratio2, E_p, f, I_s, A_1, J_s,  J_r, R_s, R_R, L_r, L_s, L_sm, Mass, \
            K_rad, Losses, gen_eff, Copper,  Iron, Structural_mass, TC1, TC2, Current_ratio, Overall_eff, cm, I

'''
        # Unpack inputs
        rad_ag               = inputs['rad_ag']
        len_s                = inputs['len_s']
        h_s                  = inputs['h_s']
        h_r                  = inputs['h_r']
        I_0                  = inputs['I_0']
        
        machine_rating       = inputs['machine_rating']
        n_nom                = inputs['n_nom']
        Gearbox_efficiency   = inputs['Gearbox_efficiency']
        
        rho_Fe               = inputs['rho_Fe']
        rho_Copper           = inputs['rho_Copper']
        
        B_symax              = inputs['B_symax']
        S_Nmax               = inputs['S_Nmax']
        
        shaft_cm             = inputs['shaft_cm']
        shaft_length         = inputs['shaft_length']


        outputs['B_g']                = B_g
        outputs['B_g1']               = B_g1
        outputs['B_rymax']            = B_rymax
        outputs['B_tsmax']            = B_tsmax
        outputs['B_trmax']            = B_trmax

        outputs['q1']                 = q1
        outputs['N_s']                = N_s
        outputs['S']                  = S
        outputs['h_ys']               = h_ys
        outputs['b_s']                = b_s
        outputs['b_t']                = b_t

        outputs['D_ratio']            = D_ratio
        outputs['A_Cuscalc']          = A_Cuscalc
        outputs['Slot_aspect_ratio1'] = Slot_aspect_ratio1
        outputs['h_yr']               = h_yr

        outputs['tau_p']              = tau_p
        outputs['p']                  = p
        outputs['Q_r']                = Q_r
        outputs['N_r']                = N_r
        outputs['b_r']                = b_r
        outputs['b_trmin']            = b_trmin

        outputs['b_tr']               = b_tr
        outputs['A_Curcalc']          = A_Curcalc
        outputs['Slot_aspect_ratio2'] = Slot_aspect_ratio2
        outputs['E_p']                = E_p
        outputs['f']                  = f
        
        outputs['I_s']                = I_s
        outputs['A_1']                = A_1
        outputs['J_s']                = J_s
        outputs['J_r']                = J_r
        outputs['R_s']                = R_s
        outputs['R_R']                = R_R

        outputs['L_r']                = L_r
        outputs['L_s']                = L_s
        outputs['L_sm']               = L_sm
        outputs['Mass']               = Mass
        outputs['K_rad']              = K_rad
        outputs['Losses']             = Losses

        outputs['gen_eff']            = gen_eff
        outputs['Copper']             = Copper
        outputs['Iron']               = Iron
        outputs['Structural_mass']    = Structural_mass
        outputs['TC1']                = TC1
        outputs['TC2']                = TC2

        outputs['Current_ratio']      = Current_ratio
        outputs['Overall_eff']        = Overall_eff
        outputs['cm']                 = cm
        outputs['I']                  =  I
'''       
        


""" ----------------
This tool was adapted from FusedWind-0.1.0 runSuite.  The interface is wrapped for better itegration with
AeroElasticSE, but the overall functionality is unchanged.
---------------- """

"""
This module allows for sampling of distributions encountered in wind energy research using a flexible 
text input file specification.  The distributions currently available are:

- Weibull
- Gamma
- vonMises
- uniform
- normal
- bivariate normal

In addition, we allow for "enumeration" of sets of variables, and their cartesian product.

The impetus is the desire for a generic representation of rows in Table 1 of IEC document 88_329
describing design standards for loads analysis, where what is being represented are the
distributions of wind/wave environment parameters.
We can assume these all boil down to distributions of _numerical_ (rather than discrete) parameters.
A twist is that some of the parameters depend on eachother, that is, the distributions are conditional distributions.

In principle all we need to say is  :math:`\{v_i\} \sim D_k (\{p_j(v)\})`
where :math:`\{v_i\}` are the random variables we're defining (including multivariate distributions),
:math:`D_k` are distributions, and :math:`\{p_j\}` are parameters to these distributions.
Currently :math:`p_j` can only depend on :math:`v_i` for :math:`i<j`, but otherwise
the dependence can be expressed as an arbitrary python math expression.

An example of the input this system can handle is::

    #--WW4Ddist--
    Vhub = W(2.120, 9.767)
    WaveDir = VM(0.029265 + 0.119759 * Vhub, -0.620138 + 0.030709 * Vhub)
    Hs = G(4.229440 + 0.222745 * WaveDir  + 0.328602 * Vhub, 0.145621 + -0.006188 * WaveDir  + 0.007561 * Vhub)
    Tp = G(19.677865 + 6.617868 * Hs  + -0.532952 * Vhub, 0.202077 + 0.286631 / Hs  + -0.004321 * Vhub)

which causes the code to sample the joint distribution of hub height wind speed (Weibull), wave direction (von Mises), 
wave height (Gamma), and wave period (Gamma) for a recent study.

Or::

    #--WW4Ddist--
    Vhub = {5,15,25}
    Hs = {1,5}
    
which causes the code to generate the cartesian product {(5,1),(5,5),(15,1),(15,5),(25,1),(25,5)}.

"""

### This is some more of the thinking that was part of development, so not included in the docstring:
#Some cases:
#x = 7
#x ~ unif({7})

#x ~ N(0,1)
#y = x
#y ~ delta(x)

#x = 7
#y = E[y|x] = \int{ p(y|x) dy }
#requires
#p(y|x), eg
#y|x ~ N(2x/7, 1)

#x,y "jointly distributed"
#x,y ~ N([0,0], [[sigmx, covxy],[covxy,sigy]])

#Given these specifications, we can then sample the distn's however we want
#ie we end up with variables like
#d = Distn()
#d.type = "gaussian"
#d.var = "x"
#d.mean = 7
#d.var = 11

#d = Distn()
#d.type = "joint_gaussian"
#d.var = ["x","y"]
#d.mean = [mux, muy]
#d.varcovar = [[etc.][]]

#or then a whole class hierarchy for distributions
#d = JointGaussian(variables, means, varcovarmatrix)
#d = JointUniform(vars,[[x1,x2][y1,y2])
#d = DiscreteUniform(var, {a,b,c})   (for enumerations)

#or special d = Enumeration(var, {a,b,c})

#There are enumerations and distributions
#Enumerations have fixed set of values returned by d.sample()
#Distributions have a number of samples returned by d.sample()

#Then we take 
#d1.sample() X d2.sample() X ...
#until all the relevant variables are specified


teststr = """
x ~ {7,8}
sigy = 1
y = N(2*x, sigy)
"""

import re, random, numpy as np, numpy.random as npr
from copy import deepcopy

from scipy.stats import vonmises, gamma
import scipy.integrate as integrate
# this is not available until scipy 0.14:
#from scipy.stats import  multivariate_normal
# so we have to us numpy.multivariate_normal (at least on my mac)
import scipy.linalg as linalg
from math import *
import math
from numpy import matrix
from math import pi, isnan, exp
from numpy import mean
from scipy.stats import vonmises, gamma

global gIgnoreJointDistn
gIgnoreJointDistn = True

# from CaseGen_General import save_case_matric, case_naming

def draw_uniform(x0,x1):
    val = npr.uniform(x0,x1)
    return val

def draw_normal(mu, sigma):            
    val = npr.normal(mu, sigma)
    return val

def draw_weibull(shape, scale, nsamples=1):
    from numpy.random import weibull
    x = scale * weibull(shape, nsamples)
    return x

def draw_gamma(shape, scale, nsamples=1):
#    print("gamma params:", shape, scale)
    shape = max(1e-3,shape)
    scale = max(1e-3,scale)
    x = gamma.rvs(shape, loc=0, scale=scale, size=nsamples)
    return x

def draw_vonmises(kappa, loc, nsamples=1):
#    print("kappa, loc: ", kappa, loc)
#    x = 180/pi * vonmises.rvs(kappa, loc=loc, size=nsamples)
    x = vonmises.rvs(kappa, loc=loc, size=nsamples)
    return x

def draw_multivariate_normal(mean, cov):
    val = npr.multivariate_normal(mean, cov)
#    val = multivariate_normal.rvs(mean, cov)
    return val

############

def prob_uniform(x,x0,x1):
    return 1/float(x1-x0)

def prob_normal(x, mu, sigma):
    val = 1/(sigma * np.sqrt(2 * np.pi)) * np.exp( - (x - mu)**2 / (2 * sigma**2))
    return val    

def prob_weibull(x, shape, scale):
    p = (shape/scale)*(x/scale)**(shape-1)*exp(-(x/scale)**shape)
    return p

def prob_gamma(x, shape, scale):
    shape = max(1e-1,shape)
    p = gamma.pdf(x,shape, loc=0, scale=scale)
    return p

def prob_vonmises(x, kappa, loc):
#    print("kappa, loc: ", kappa, loc)
#    p = vonmises.pdf(x*pi/180.0,kappa, loc=loc)
#    p = [vonmises.pdf(x*pi/180.0 + m,kappa, loc=loc) for m in [-2*pi, 0, 2*pi]]
    p = [vonmises.pdf(x + m,kappa, loc=loc) for m in [-2*pi, 0, 2*pi]]
    p = max(p)
    return p

def prob_multivariate_normal(x, mu, sigma):
## not in scipy < 0.14:
#    p = multivariate_normal.pdf(x,mean,cov)
#    return p
# instead:
    x = np.array(x)
    mu = np.array(mu)
    sigma = matrix(sigma)
    size = len(x)
    if size == len(mu) and (size, size) == sigma.shape:
        det = linalg.det(sigma)
        if det == 0:
            raise NameError("The covariance matrix can't be singular")
        norm_const = 1.0/ ( math.pow((2*pi),float(size)/2) * math.pow(det,1.0/2) )
        x_mu = matrix(x - mu)
        inv = sigma.I        
        result = math.pow(math.e, -0.5 * (x_mu * inv * x_mu.T))
        return norm_const * result
    else:
        raise NameError("The dimensions of the input don't match")


###############################
###############################

#def math_op(op,v1,v2):
#    if (op == "+"):
#        return v1+v2
#    elif (op == "-"):
#        return v1-v2
#    elif (op == "*"):
#        return v1*v2
#    elif (op == "/"):
#        return v1/v2
#    else:
#        raise ValueError("unknown math op %s" % op)

def is_float(s):
    try:
        v = float(s)
        return True
    except:
        return False


def parse_arg(a):
    from shlex import shlex
    obj = shlex(a)
    obj.wordchars += "." # enables parsing decimals
    alist = list(obj)
#    print("parsed ", a, "to ", alist)
#eg: parsed  10 + Vhub/100 - 0.02*(WaveDir+1) to  ['10', '+', 'Vhub', '/', '100', '-', '0.02', '*', '(', 'WaveDir', '+', '1', ')']
    return alist

def merge_dicts(d1,d2):
    d =  {key:d1[key] for key in d1}
    for key in d2:
        d[key] = d2[key]
    return d


class Distribution(object):
    def __init__(self, vstr, ctx):
        self.vstr = vstr # what variable is being defined
        self.ctx = ctx

    def get_bounds(self):
        pass
    def get_partitions(self):
        pass
    def get_name(self):
        return self.vstr
    def calc_prob(self, x):
        return 1
    
    
class EnumDistn(Distribution):
    """ class for distributions that are just list of numbers or single numbers,
    'sampling' is just going to return all values"""
    def __init__(self, vstr, items, ctx):
        super(EnumDistn, self).__init__(vstr, ctx)
        self.items = items

    def sample(self):
#        res = []
#        for i in range(nsample):
#        res.append(random.choice(self.items))
        res = random.choice(self.items)
        res = self.ctx.resolve_value(res)
        return res

    def get_bounds(self):
        """ assumes numbers """
        nums = [float(i[0]) for i in self.items]
        return [min(nums), max(nums)]

    def get_partitions(self):
        return (len(self.items) - 1)
        
    

class FnDistn(Distribution):
    """ class for 'real' distributions like Gaussians, etc. 
    the keys in 'items' are the names of the distributions, assumed to be known 
    and sample-able. items = e.g. {"N": ["2x","sig1"} """
    def __init__(self, vstr, fn, args, ctx):
        super(FnDistn, self).__init__(vstr,ctx)
        self.fn = fn
        self.args = args

        self.is_truncated = False
        self.C = None
        if (fn[0] == "T"):  ## this is a truncated distribution
            self.is_truncated = True
            self.fn = fn[1:]  # strip "T"            
            print("distn is truncated", fn, self.fn)

    def ensure_C(self, argvals):
        if not self.is_truncated:
            return
        if self.C == None:
            self.Vin = argvals[-2]  # last two args to distn are truncation points
            self.Vout = argvals[-1]
            self.is_truncated = False  # temp turn it off b/c we want true integral of pdf over truncated range
            self.C = integrate.quad(self.calc_prob, self.Vin, self.Vout)
            self.is_truncated = True
            self.C = self.C[0]
            print("truncated distribution normalization constant = ", self.C, "range: ", self.Vin, self.Vout)

    def sample(self):
        argvals = []
        for i in range(len(self.args)):
            a = self.args[i]
            argvals.append(self.ctx.resolve_value(a))

        if (self.is_truncated):
            self.ensure_C(argvals)
            val = self.Vin - 1
            while (val < self.Vin or val > self.Vout):
                val = self.raw_sample(argvals)
            return val
        else:
            return self.raw_sample(argvals)

    def raw_sample(self, argvals):
        if (self.fn == "N"):
#            print(" need to sample normal with args = ", argvals)
            val = draw_normal(argvals[0], argvals[1])
        elif (self.fn == "N2"):
#            print(" need to sample 2D normal with args = ", argvals)
            val = draw_multivariate_normal([argvals[0], argvals[1]], [[argvals[2], argvals[4]],[argvals[4], argvals[3]]])
        elif (self.fn == "U"):
#            print(" need to sample uniform with args = ", argvals)
            val = draw_uniform(argvals[0], argvals[1])
        elif (self.fn == "G"):
#            print(" need to sample gamma with args = ", argvals)
            val = draw_gamma(argvals[0], argvals[1])
        elif (self.fn == "VM"):
#            print(" need to sample von Mises with args = ", argvals)
            val = draw_vonmises(argvals[0], argvals[1])
        elif (self.fn == "W"):
            val = draw_weibull(argvals[0], argvals[1])
#            print(" sampled Wiebull with args = ", argvals, "got ", val)
        else:
            raise ValueError("Sorry, unknown distribution: %s" % self.fn)
        return val

    def calc_prob(self, x):
#        print("calc_prob", self.fn, x)
#        x = x[0]
        argvals = []
        for i in range(len(self.args)):
            a = self.args[i]
            argvals.append(self.ctx.resolve_value(a))

        if (self.is_truncated):
            self.ensure_C(argvals)
            if x < self.Vin or x > self.Vout:
                return 0  ## bail here for samples outside range
            C = self.C ## this is normalization for truncated distn
        else:
            C = 1.0  ## untruncated distn's have normalization of 1
            
        if (self.fn == "N"):
            val = prob_normal(x,argvals[0], argvals[1])
        elif (self.fn == "N2"):
            val = prob_multivariate_normal(x,[argvals[0], argvals[1]], [[argvals[2], argvals[4]],[argvals[4], argvals[3]]])
        elif (self.fn == "U"):
            val = prob_uniform(x,argvals[0], argvals[1])
        elif (self.fn == "G"):
            val = prob_gamma(x,argvals[0], argvals[1])
        elif (self.fn == "VM"):
            val = prob_vonmises(x,argvals[0], argvals[1])
        elif (self.fn == "W"):
            val = prob_weibull(x,argvals[0], argvals[1])
#            print("prob W", x, argvals[0], argvals[1], val)
        else:
            raise ValueError("unknown distribution %s" % self.fn)
        
        if (math.isnan(val)):
            print("NAN", val, x, self.fn, argvals)
        return val/C


class DistnParser(object):
    def __init__(self):
        self.vars = []
        self.dlist = []
        self.dlist_map = {}
        
    def parse_file(self,fname):
        mystr = file(fname).readlines()
        return self.parse(mystr)

    def parse(self, mystr):
        for ln in mystr:
            ln = ln.strip()
            newdist = None
            if (len(ln) > 0 and ln[0] != "#"):
                # first get rid of anything past any "#" on the right
                ln = ln.split("#")[0]
 #               print("your line: ", ln)
                # separate vars from the distn
                tok = ln.split("=")
                if len(tok) > 1:
#                    print("found distn: ", tok)
                    vtok = tok[0].split(",")
                    # parse out the vars in question:
                    vstr = ""
                    for v in vtok:
                        v = v.strip()
                        if (v in self.vars):
                            print("ERROR: Variable %s is doubly defined" %v)
                        self.vars.append(v)
                        vstr = "%s" % v
                    # now parse the distn spec.
                    dspec = tok[1].strip()
#                    print("look at", dspec)
                    q = re.match("([^(]+)(\(.*\))", dspec)
                    if (q != None):
                        # found a function-like defn
                        dist = q.group(1) 
                        args = q.group(2)
#                        print("split ", dspec , "into ", dist, args)
                        #eg: split  G(10 + Vhub/100 - 0.02*(WaveDir+1),.25) into  G (10 + Vhub/100 - 0.02*(WaveDir+1),.25)
                        args=args.strip("(").strip(")").split(",")
                        args = [s.strip() for s in args]
                        # now parse variables out of expressions
                        # simple math allowed: <val> := <number> | <val> <op> <val> where <op> = +-*/
                        arglist = []
                        for a in args:
                            alist = parse_arg(a)
                            arglist.append(alist)    
#                        print("found dist=", dist, " arglist=", arglist)
                        newdist = FnDistn(vstr,dist,arglist, self)
                        self.dlist.append(newdist)
                    elif dspec[0] == "{":
                        # found a set
                        args = [s.strip() for s in dspec.strip("{").strip("}").split(",")]
#                        args = [float(s) for s in args]
#                        print("found set ", args)
                        args = [parse_arg(a) for a in args]
#                        print("FOUND set ", args)
                        newdist = EnumDistn(vstr,args, self)
                        self.dlist.append(newdist)
                    else:
                        try:
                            arg = [parse_arg(dspec)]
                            # found a value. TODO: tuples
#                            print("found number ", arg)
                            newdist = EnumDistn(vstr, arg, self)
                            self.dlist.append(newdist)
                        except:
                            print("cannot parse distribution spec ", dspec)
                if (newdist != None):
                    self.dlist_map[vstr] = newdist
                        
        print("defined distns. for vars ", self.vars)
#        print(self.dlist       )
#        print("dlist map = ", self.dlist_map)

        return self.dlist

    def clear_values(self,):
        self.values = {}
    def set_value(self, s,v):
        global gIgnoreJointDistn
        self.values[s] = v
        if (not gIgnoreJointDistn):
            # also deconstruct joint distribution values into their individual variables
            # e.g. Hs.Tp: [1,3] -> Hs: 1, Tp:3
            subv = s.split(".")
            if (len(subv) > 1):
                for i in range(len(subv)):
                    self.values[subv[i]] = v[i]

    def set_values(self, e):
        for key in e:
            self.set_value(key, e[key])

    def sample(self):
        self.clear_values()
        for d in self.dlist:
            s = d.sample()
#            print("sampled ", d.vstr, " and got ", s)
            self.set_value(d.vstr,s)

    def add_enum(self, slist, enum):
        """
        slist = list of samples we're building
        enum = distn with enum.items list of possible values
        output = slist X enum.items
        """
        newlist = []
#        print("add enum, slist, enum.items", slist, enum.items)
        for s in slist: 
            for x in enum.items:
                item = {}
                for y in s:
                    item[y] = s[y]
                    self.set_value(y,s[y])
#                item[enum.vstr] = self.resolve_value(x)
                item[enum.vstr] = x  # not resolved yet!
                newlist.append(item)                
#                print("appended item to newlist" , item, newlist)
#        print(newlist)
        return newlist

    def expand_enums(self):
        """ return list of dicts """
        self.clear_values()
        slist = [{}]
        for d in self.dlist:
            if (hasattr(d,"items")):  #### bad programming!
                slist = self.add_enum(deepcopy(slist), d)
        return slist

    def sample_fns(self):
        """ return list of dicts """
        for d in self.dlist:
            if (hasattr(d,"fn")):
                s = d.sample()
#                print("sampled ", d.vstr, " and got ", s)
                self.set_value(d.vstr,s)
        return self.values

    def multi_sample(self, numsamples, expand_enums=False):
        slist = []
        if (expand_enums):
#            print("expanding set variables, then sampling %d times" % numsamples)
            # make slist of product space of set vars.
            enum_list = self.expand_enums()
#            print("enum_list", enum_list)
            for e in enum_list:
#                self.clear_values()
#                self.set_values(e)
                for i in range(numsamples):
                    # now truly sampled vars to each of them                    
                    # combine real samples to enum cases
                    self.clear_values()

                    for d in self.dlist:
 #                       print("scanning ", d.vstr)
                        if (hasattr(d,"fn")):
                            s = d.sample()
                            #print("sampled ", d.vstr, " and got ", s)
                            self.set_value(d.vstr,s)
                        
                        if (hasattr(d,"items")):  #### bad programming!
                            maxiter = len(e)
                            tries = 0
                            while (tries < maxiter):
                                for it in e:  # find this var and resolve it here
#                                    print(it, e[it])
                                    # this is some crazy stuff: try (maybe out of order b/c of dict) until we succeed
                                    # to resolve everything
                                    try:
                                        val = self.resolve_value(e[it])
                                        self.set_value(it,val)
#                                        print("succes resolving ", e[it])
                                    except:
#                                        print("failed resolving ", e[it])
                                        pass
                                    tries += 1
                    slist.append(self.values)

#                    vals = self.sample_fns()
#                    print(vals, e)
#                    vals = merge_dicts(vals, e)
#                    slist.append(vals)
        else:            
#            print("sampling %d times" % numsamples)
            for i in range(numsamples):
                self.clear_values()
                vals = self.sample()
#                print("sample %d = " % i, self.values)
                slist.append(self.values)
        
        return slist

    def resolve_one_value(self,a):
#        print("resolve_one_value()", a)
        if (is_float(a)):
            return float(a)
        else:
            for d in self.dlist:
                if a == d.vstr:
                    return self.values[a]
#            raise ValueError("did not find variable:%s: in dlist" % a )
        return a

    def resolve_value(self,a):
#        print("resolving:", a)
        vals = [self.resolve_one_value(x) for x in a]
#        print(vals)
        s = ""
        for v in vals:
            if is_float(v):
                s = "%s %f" % (s, v)
            else:
                s = "%s %s" % (s, v)
#        print("evaluating:", s)
        val = eval(s)
#        print("= ", val)
        return val
#        if (len(a) == 1):
#            return self.resolve_one_value(a[0])
#        else:
#            v1 = self.resolve_one_value(a[0])
#            v2 = self.resolve_one_value(a[2])
#            op = a[1]
#            print("2",a)
#            v1 = self.resolve_one_value(a[-1])
#            v2 = self.resolve_value(a[:-2])
#            op = a[-2]
#            return math_op(op, v1, v2)

    def get_num_samples(self):
        self.sample()  # kludge: sample once to get NumSamples set, if it exists
        if "NumSamples" in self.values:
            return int(self.values['NumSamples'])
        else:
            return 1

    def get_bounds(self):
        """ assume all enum distns, get their bounds for conversion to dakota """
        bounds = []
        for d in self.dlist:
            bounds.append(d.get_bounds())
        low = [b[0] for b in bounds]
        high = [b[1] for b in bounds]
        return [low,high]

    def get_partitions(self):
        """ assume all enum, get their length, assume uniform incrs to convert to dakota """
        part = []
        for d in self.dlist:
            part.append(d.get_partitions())
        return part

    def get_names(self):
        """ get names of all the distn's (what vars are defined), for mapping from dakota variable list """
        part = []
        for d in self.dlist:
            part.append(d.get_name())
        return part

    def calc_prob(self, samp):
        """ calc probability of samp according to parsed distribution """
        self.clear_values()
        self.set_values(samp)        
        # for each key/value, need to find the distribution it belongs to
        # plan on using a pre-prepared mapping
        ptot = 1
#        print("calc prob for ", samp)
        for var in self.values:
            if (var in self.dlist_map):
                dist = self.dlist_map[var]
#                print("for ", var)
                p = dist.calc_prob(self.values[var])
                ptot *= p
        return ptot


# def get_options():
#     from optparse import OptionParser
#     parser = OptionParser()    
#     parser.add_option("-i", "--input", dest="dist",  type="string", default="runbatch-dist.txt",
#                                     help="main input file describing distribution, ie cases to run")
#     parser.add_option("-n", "--nsamples", dest="nsamples", help="how many samples to generate", type="int", default=5)
#     parser.add_option("-t", "--tmax", dest="tmax", help="analysis time, default is use tag in distribution file", type="float", default=None)
#     parser.add_option("-o", "--output", dest="main_output",  type="string", default="runcases.txt",
#                                     help="output file (where to write the run cases)")
#     parser.add_option("-p", "--probfile", dest="old_samples",  type="string", default=None,
#                                     help="an input file of samples whose probabilities we want to calculat w.r.t input distn")
#     parser.add_option("-a", "--augment", dest="augment", help="goes with -p, will include _both_ given and newly calculated probs to the output samples", action="store_true", default=False)
            
#     (options, args) = parser.parse_args()
#     return options, args

def read_samples(fname):
    lines = file(fname).readlines()
    hdr = lines[0].split()
    dat = []
    for ln in lines[1:]:
        dat.append([float(x) for x in ln.split()])
    return hdr, dat

def gen_cases(options=None, args=None):
    if options==None:
        options, args = get_options()
    
    dparser = DistnParser()
    dparser.parse_file(options['dist'])

    if (options['old_samples'] != None):
        # in this mode, we are given a distribution and, separately, a set of old samples.  Our job is to calculate the
        # probabilities for the samples w.r.t. the given distribution
        old_hdr, old_samples = read_samples(options['old_samples'])
        pidx = old_hdr.index("Prob")
        new_hdr = old_hdr
        if (options['augment']):
            new_hdr.append("Prob2")
            pidx = len(new_hdr)-1
        new_samples = []
        for s in old_samples:
            samp = {old_hdr[i]:s[i] for i in range(len(s))}
            p = dparser.calc_prob(samp)
            if (options['augment']):
                s.append(p)
            else:
                s[pidx] = p
            new_samples.append(s)

        fout = file(options['main_output'], "w")
        for key in new_hdr:
            fout.write("%s " % key)
        fout.write("\n")

        for s in new_samples:
            for val in s:
                fout.write("%.16e " % val)
            fout.write("\n")
        fout.close()

        print("Calculated probabilities of samples in %s w.r.t. distribution in %s" % (options['old_samples'], options['dist']))
        if (options['augment']):
            print("Augmented output with these probs as Prob2 in field (0-based) %d" % (pidx))
        else:
            print("Replaced old probs with new probs in output")

    else:
        numsamples = options['nsamples']
        slist = dparser.multi_sample(numsamples, expand_enums = True)
        case_list_dist = [{}]*numsamples
        for i in range(len(slist)):
            s = slist[i]
            p =  dparser.calc_prob(s)

            case_list_dist[i] = s
            case_list_dist[i]['p'] = p

        return case_list_dist


def CaseGen_Dists(case_inputs):

    # Default options
    if 'dist' not in case_inputs.keys():
        case_inputs['dist'] = 'runbatch-dist.txt'
    if 'augment' not in case_inputs.keys():
        case_inputs['augment'] = False
    if 'tmax' not in case_inputs.keys():
        case_inputs['tmax'] = 180
    if 'old_samples' not in case_inputs.keys():
        case_inputs['old_samples'] = None
    if 'nsamples' not in case_inputs.keys():
        case_inputs['nsamples'] = 5
    if 'main_output' not in case_inputs.keys():
        case_inputs['main_output'] = 'run_cases.txt'
    if 'dir_matrix' not in case_inputs.keys():
        case_inputs['dir_matrix'] = ''
    if 'namebase' not in case_inputs.keys():
        case_inputs['namebase'] = ''

    # Read distribution input file, generate cases
    case_list_dist = gen_cases(case_inputs, [])

    # Reformat for AeroElasticSE case runner
    case_list = [] 
    for i, case in enumerate(case_list_dist):
        case_list_i = {}
        if 'AnalTime' in case.keys():
            case_list_i[('Fst','TMax')] = [case['AnalTime']]
        if 'Hs' in case.keys():
            case_list_i[('HydroDyn','WaveHs')] = case['Hs']
        if 'Tp' in case.keys():
            case_list_i[('HydroDyn','WaveTp')] = case['Tp']
        if 'WaveDir' in case.keys():
            case_list_i[('HydroDyn','WaveDir')] = case['WaveDir']
        if 'Vhub' in case.keys():
            case_list_i[('InflowWind','HWindSpeed')] = case['Vhub']
        if 'p' in case.keys():
            case_list_i[('P','')] = case['p']

        case_list.append(case_list_i)

    # Save case matrix
    change_vars = sorted(case_list[0].keys())
    matrix_out = np.asarray([[str(case[var][0]) for var in change_vars] for case in case_list])
    if case_inputs['dir_matrix']:
        save_case_matric(matrix_out, change_vars, case_inputs['dir_matrix'])

    # Case naming
    case_name = case_naming(len(case_list), namebase=case_inputs['namebase'])

    return case_list, case_name


if __name__=="__main__":

    case_inputs = {}
    case_inputs['dist']        = 'runbatch-dist.txt'
    case_inputs['tmax']        = 30
    case_inputs['nsamples']    = 5
    case_inputs['main_output'] = 'run_cases.txt'
    case_inputs['dir_matrix']  = 'temp/openFAST'
    case_inputs['namebase']    = 'testing_dists'

    case_list, case_name = CaseGen_Dists(case_inputs)




from __future__ import print_function
import numpy as np
import warnings
import copy, time

from openmdao.api import ExplicitComponent, Group, IndepVarComp, Problem

from wisdem.commonse.akima import Akima, akima_interp_with_derivs
from wisdem.ccblade.ccblade_component import CCBladeGeometry
from wisdem.ccblade import CCAirfoil
from wisdem.airfoilprep import Airfoil
from wisdem.rotorse.rotor_geometry_yaml import ReferenceBlade
from wisdem.rotorse.precomp import PreComp, Profile, Orthotropic2DMaterial, CompositeSection, _precomp
from wisdem.rotorse.rotor_cost import blade_cost_model

from scipy.interpolate import PchipInterpolator

NINPUT = 8



class BladeGeometry(ExplicitComponent):
    def initialize(self):
        self.options.declare('RefBlade')

        # Blade Cost Model Options
        self.options.declare('verbosity',           default=False)
        self.options.declare('tex_table',           default=False)
        self.options.declare('generate_plots',      default=False)
        self.options.declare('show_plots',          default=False)
        self.options.declare('show_warnings',       default=False)
        self.options.declare('discrete',            default=False)
        self.options.declare('user_update_routine', default=None)
    
    def setup(self):
        self.refBlade = RefBlade = self.options['RefBlade']
        npts    = len(self.refBlade['pf']['s'])
        NINPUT  = len(self.refBlade['ctrl_pts']['r_in'])
        NAF     = len(self.refBlade['outer_shape_bem']['airfoil_position']['grid'])
        NAFgrid = len(self.refBlade['airfoils_aoa'])
        NRe     = len(self.refBlade['airfoils_Re'])

        self.add_input('bladeLength',   val=0.0, units='m', desc='blade length (if not precurved or swept) otherwise length of blade before curvature')
        self.add_input('r_max_chord',   val=0.0, desc='location of max chord on unit radius')
        self.add_input('chord_in',      val=np.zeros(NINPUT), units='m', desc='chord at control points')  # defined at hub, then at linearly spaced locations from r_max_chord to tip
        self.add_input('theta_in',      val=np.zeros(NINPUT), units='deg', desc='twist at control points')  # defined at linearly spaced locations from r[idx_cylinder] to tip
        self.add_input('precurve_in',   val=np.zeros(NINPUT), units='m', desc='precurve at control points')  # defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
        self.add_input('presweep_in',   val=np.zeros(NINPUT), units='m', desc='precurve at control points')  # defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
        self.add_input('sparT_in',      val=np.zeros(NINPUT), units='m', desc='thickness values of spar cap that linearly vary from non-cylinder position to tip')
        self.add_input('teT_in',        val=np.zeros(NINPUT), units='m', desc='thickness values of trailing edge panels that linearly vary from non-cylinder position to tip')
        # self.add_input('leT_in',        val=np.zeros(NINPUT), units='m', desc='thickness values of leading edge panels that linearly vary from non-cylinder position to tip')
        self.add_input('airfoil_position', val=np.zeros(NAF), desc='spanwise position of airfoils')

        # parameters
        self.add_input('hubFraction', val=0.0, desc='hub location as fraction of radius')

        # Blade geometry outputs
        self.add_output('Rhub', val=0.0, units='m', desc='dimensional radius of hub')
        self.add_output('Rtip', val=0.0, units='m', desc='dimensional radius of tip')
        self.add_output('r',            val=np.zeros(npts), units='m', desc='dimensional aerodynamic grid')
        self.add_output('r_in',         val=np.zeros(NINPUT), units='m', desc='Spline control points for inputs')
        self.add_output('max_chord',    val=0.0, units='m', desc='maximum chord length')
        self.add_output('chord',        val=np.zeros(npts), units='m', desc='chord at airfoil locations')
        self.add_output('theta',        val=np.zeros(npts), units='deg', desc='twist at airfoil locations')
        self.add_output('precurve',     val=np.zeros(npts), units='m', desc='precurve at airfoil locations')
        self.add_output('presweep',     val=np.zeros(npts), units='m', desc='presweep at structural locations')
        self.add_output('rthick',       val=np.zeros(npts), desc='relative thickness of airfoil distribution')
        self.add_output('le_location',  val=np.zeros(npts))
        self.add_output('hub_diameter', val=0.0, units='m')
        self.add_output('diameter',     val=0.0, units='m')

        # Airfoil properties
        # self.add_discrete_output('airfoils', val=[], desc='Spanwise coordinates for aerodynamic analysis')
        self.add_output('airfoils_cl',  val=np.zeros((NAFgrid, npts, NRe)), desc='lift coefficients, spanwise')
        self.add_output('airfoils_cd',  val=np.zeros((NAFgrid, npts, NRe)), desc='drag coefficients, spanwise')
        self.add_output('airfoils_cm',  val=np.zeros((NAFgrid, npts, NRe)), desc='moment coefficients, spanwise')
        self.add_output('airfoils_aoa', val=np.zeros((NAFgrid)), units='deg', desc='angle of attack grid for polars')
        self.add_output('airfoils_Re',  val=np.zeros((NRe)), desc='Reynolds numbers of polars')
        
        # Airfoil coordinates
        self.add_output('airfoils_coord_x',  val=np.zeros((200, npts)), desc='x airfoil coordinate, spanwise')
        self.add_output('airfoils_coord_y',  val=np.zeros((200, npts)), desc='y airfoil coordinate, spanwise')
        
        # Beam properties
        self.add_output('z',            val=np.zeros(npts), units='m',      desc='locations of properties along beam')
        self.add_output('EA',           val=np.zeros(npts), units='N',      desc='axial stiffness')
        self.add_output('EIxx',         val=np.zeros(npts), units='N*m**2', desc='edgewise stiffness (bending about :ref:`x-direction of airfoil aligned coordinate system <blade_airfoil_coord>`)')
        self.add_output('EIyy',         val=np.zeros(npts), units='N*m**2', desc='flatwise stiffness (bending about y-direction of airfoil aligned coordinate system)')
        self.add_output('EIxy',         val=np.zeros(npts), units='N*m**2', desc='coupled flap-edge stiffness')
        self.add_output('GJ',           val=np.zeros(npts), units='N*m**2', desc='torsional stiffness (about axial z-direction of airfoil aligned coordinate system)')
        self.add_output('rhoA',         val=np.zeros(npts), units='kg/m',   desc='mass per unit length')
        self.add_output('rhoJ',         val=np.zeros(npts), units='kg*m',   desc='polar mass moment of inertia per unit length')
        self.add_output('Tw_iner',      val=np.zeros(npts), units='m',      desc='y-distance to elastic center from point about which above structural properties are computed')
        self.add_output('x_ec',         val=np.zeros(npts), units='m',      desc='x-distance to elastic center from point about which above structural properties are computed (airfoil aligned coordinate system)')
        self.add_output('y_ec',         val=np.zeros(npts), units='m',      desc='y-distance to elastic center from point about which above structural properties are computed')
        self.add_output('flap_iner',    val=np.zeros(npts), units='kg/m',   desc='Section flap inertia about the Y_G axis per unit length.')
        self.add_output('edge_iner',    val=np.zeros(npts), units='kg/m',   desc='Section lag inertia about the X_G axis per unit length')
        self.add_output('eps_crit_spar',    val=np.zeros(npts), desc='critical strain in spar from panel buckling calculation')
        self.add_output('eps_crit_te',      val=np.zeros(npts), desc='critical strain in trailing-edge panels from panel buckling calculation')
        self.add_output('xu_strain_spar',   val=np.zeros(npts), desc='x-position of midpoint of spar cap on upper surface for strain calculation')
        self.add_output('xl_strain_spar',   val=np.zeros(npts), desc='x-position of midpoint of spar cap on lower surface for strain calculation')
        self.add_output('yu_strain_spar',   val=np.zeros(npts), desc='y-position of midpoint of spar cap on upper surface for strain calculation')
        self.add_output('yl_strain_spar',   val=np.zeros(npts), desc='y-position of midpoint of spar cap on lower surface for strain calculation')
        self.add_output('xu_strain_te',     val=np.zeros(npts), desc='x-position of midpoint of trailing-edge panel on upper surface for strain calculation')
        self.add_output('xl_strain_te',     val=np.zeros(npts), desc='x-position of midpoint of trailing-edge panel on lower surface for strain calculation')
        self.add_output('yu_strain_te',     val=np.zeros(npts), desc='y-position of midpoint of trailing-edge panel on upper surface for strain calculation')
        self.add_output('yl_strain_te',     val=np.zeros(npts), desc='y-position of midpoint of trailing-edge panel on lower surface for strain calculation')

        # Blade Cost
        self.add_output('total_blade_cost', val=0.0, units='USD', desc='total blade cost')
        self.add_output('total_blade_mass', val=0.0, units='USD', desc='total blade cost')

        #
        self.add_discrete_output('blade_out', val={}, desc='updated blade dictionary for ontology')
        self.add_output('sparT_in_out', val=np.zeros(NINPUT), units='m', desc='thickness values of spar cap that linearly vary from non-cylinder position to tip, pass through for nested optimization')
        
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        # initialization point
        # if discrete_inputs['blade_in_overwrite'] != {}:
        #     blade = copy.deepcopy(discrete_inputs['blade_in_overwrite'])
        # else:
        blade = copy.deepcopy(self.refBlade)
        NINPUT = len(blade['ctrl_pts']['r_in'])

        # Set inputs to update blade geometry
        Rhub = inputs['hubFraction'] * inputs['bladeLength']
        Rtip = Rhub + inputs['bladeLength']

        outputs['Rhub']     = Rhub
        outputs['Rtip']     = Rtip
        
        r_in = blade['ctrl_pts']['r_in']
        
        outputs['r_in'] = Rhub + (Rtip-Rhub)*np.array(r_in)

        blade['ctrl_pts']['bladeLength']  = inputs['bladeLength']
        blade['ctrl_pts']['r_in']         = r_in
        blade['ctrl_pts']['chord_in']     = inputs['chord_in']
        blade['ctrl_pts']['theta_in']     = inputs['theta_in']
        blade['ctrl_pts']['precurve_in']  = inputs['precurve_in']
        blade['ctrl_pts']['presweep_in']  = inputs['presweep_in']
        blade['ctrl_pts']['sparT_in']     = inputs['sparT_in']
        blade['ctrl_pts']['teT_in']       = inputs['teT_in']
        # blade['ctrl_pts']['leT_in']       = inputs['leT_in']
        blade['ctrl_pts']['r_max_chord']  = inputs['r_max_chord']

        #check that airfoil positions are increasing        
        correct_af_position = False
        airfoil_position = copy.deepcopy(inputs['airfoil_position']).tolist()
        for i in reversed(range(1,len(airfoil_position))):
            if airfoil_position[i] <= airfoil_position[i-1]:
                airfoil_position[i-1] = airfoil_position[i] - 0.001
                correct_af_position = True        
        if correct_af_position:
            blade['outer_shape_bem']['airfoil_position']['grid'] = airfoil_position
            warning_corrected_airfoil_position = "Airfoil spanwise positions must be increasing.  Changed from: %s to: %s" % (inputs['airfoil_position'].tolist(), airfoil_position)
            warnings.warn(warning_corrected_airfoil_position)
        else:
            blade['outer_shape_bem']['airfoil_position']['grid'] = inputs['airfoil_position'].tolist()
        
        # Update
        refBlade = ReferenceBlade()
        refBlade.verbose             = False
        refBlade.NINPUT              = len(outputs['r_in'])
        refBlade.NPTS                = len(blade['pf']['s'])
        refBlade.analysis_level      = blade['analysis_level']
        refBlade.apply_stall_delay   = blade['apply_stall_delay']
        refBlade.user_update_routine = self.options['user_update_routine']
        if blade['analysis_level'] < 3:
            refBlade.spar_var        = blade['precomp']['spar_var']
            refBlade.te_var          = blade['precomp']['te_var']
            # if 'le_var' in blade['precomp']:
            #     refBlade.le_var     = blade['precomp']['le_var']
        
        blade_out = refBlade.update(blade)
        
        # Get geometric outputs
        outputs['hub_diameter'] = 2.0*Rhub
        outputs['r']            = Rhub + (Rtip-Rhub)*np.array(blade_out['pf']['s'])
        outputs['diameter']     = 2.0*outputs['r'][-1]

        outputs['chord']        = blade_out['pf']['chord']
        outputs['max_chord']    = max(blade_out['pf']['chord'])
        outputs['theta']        = blade_out['pf']['theta']
        outputs['precurve']     = blade_out['pf']['precurve']
        outputs['presweep']     = blade_out['pf']['presweep']
        outputs['rthick']       = blade_out['pf']['rthick']
        outputs['le_location']  = blade_out['pf']['p_le']

        # airfoils  = blade_out['airfoils']
        outputs['airfoils_cl']  = blade_out['airfoils_cl']
        outputs['airfoils_cd']  = blade_out['airfoils_cd']
        outputs['airfoils_cm']  = blade_out['airfoils_cm']
        outputs['airfoils_aoa'] = blade_out['airfoils_aoa']
        outputs['airfoils_Re']  = blade_out['airfoils_Re']
        
        
        
        upperCS   = blade_out['precomp']['upperCS']
        lowerCS   = blade_out['precomp']['lowerCS']
        websCS    = blade_out['precomp']['websCS']
        profile   = blade_out['precomp']['profile']
        materials = blade_out['precomp']['materials']
        
        for i in range(len(profile)):
            outputs['airfoils_coord_x'][:,i] = blade_out['profile'][:,0,i]
            outputs['airfoils_coord_y'][:,i] = blade_out['profile'][:,1,i]
        
        # Assumptions:
        # - if the composite layer is divided into multiple regions (i.e. if the spar cap is split into 3 regions due to the web locations),
        #   the middle region is selected with int(n_reg/2), note for an even number of regions, this rounds up
        sector_idx_strain_spar_ss = blade_out['precomp']['sector_idx_strain_spar_ss']
        sector_idx_strain_spar_ps = blade_out['precomp']['sector_idx_strain_spar_ps']
        sector_idx_strain_te_ss   = blade_out['precomp']['sector_idx_strain_te_ss']
        sector_idx_strain_te_ps   = blade_out['precomp']['sector_idx_strain_te_ps']


        # Get Beam Properties        
        beam = PreComp(outputs['r'], outputs['chord'], outputs['theta'], outputs['le_location'], 
                       outputs['precurve'], outputs['presweep'], profile, materials, upperCS, lowerCS, websCS, 
                       sector_idx_strain_spar_ps, sector_idx_strain_spar_ss, sector_idx_strain_te_ps, sector_idx_strain_te_ss)
        EIxx, EIyy, GJ, EA, EIxy, x_ec, y_ec, rhoA, rhoJ, Tw_iner, flap_iner, edge_iner = beam.sectionProperties()

        outputs['eps_crit_spar'] = beam.panelBucklingStrain(sector_idx_strain_spar_ss)
        outputs['eps_crit_te'] = beam.panelBucklingStrain(sector_idx_strain_te_ss)

        xu_strain_spar, xl_strain_spar, yu_strain_spar, yl_strain_spar = beam.criticalStrainLocations(sector_idx_strain_spar_ss, sector_idx_strain_spar_ps)
        xu_strain_te, xl_strain_te, yu_strain_te, yl_strain_te = beam.criticalStrainLocations(sector_idx_strain_te_ss, sector_idx_strain_te_ps)
        
        outputs['z']         = outputs['r']
        outputs['EIxx']      = EIxx
        outputs['EIyy']      = EIyy
        outputs['GJ']        = GJ
        outputs['EA']        = EA
        outputs['EIxy']      = EIxy
        outputs['x_ec']      = x_ec
        outputs['y_ec']      = y_ec
        outputs['rhoA']      = rhoA
        outputs['rhoJ']      = rhoJ
        outputs['Tw_iner']   = Tw_iner
        outputs['flap_iner'] = flap_iner
        outputs['edge_iner'] = edge_iner

        outputs['xu_strain_spar'] = xu_strain_spar
        outputs['xl_strain_spar'] = xl_strain_spar
        outputs['yu_strain_spar'] = yu_strain_spar
        outputs['yl_strain_spar'] = yl_strain_spar
        outputs['xu_strain_te']   = xu_strain_te
        outputs['xl_strain_te']   = xl_strain_te
        outputs['yu_strain_te']   = yu_strain_te
        outputs['yl_strain_te']   = yl_strain_te
        

        # Blade cost model
        bcm             = blade_cost_model(options=self.options) # <------------- options, import blade cost model
        bcm.name        = blade_out['config']['name']
        bcm.materials   = materials
        bcm.upperCS     = upperCS
        bcm.lowerCS     = lowerCS
        bcm.websCS      = websCS
        bcm.profile     = profile
        bcm.chord       = outputs['chord']
        bcm.r           = (outputs['r'] - outputs['Rhub'])/(outputs['Rtip'] - outputs['Rhub']) * float(inputs['bladeLength'])
        bcm.bladeLength = float(inputs['bladeLength'])
        bcm.le_location              = outputs['le_location']
        blade_cost, blade_mass       = bcm.execute_blade_cost_model()

        outputs['total_blade_cost'] = blade_cost
        outputs['total_blade_mass'] = blade_mass

        #
        discrete_outputs['blade_out'] = blade_out
        outputs['sparT_in_out'] = inputs['sparT_in']
        
class Location(ExplicitComponent):
    def setup(self):
        self.add_input('hub_height', val=0.0, units='m', desc='Tower top hub height')
        self.add_output('wind_zvec', val=np.zeros(1), units='m', desc='Tower top hub height as vector')
        #self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        outputs['wind_zvec'] = np.array([ np.float(inputs['hub_height']) ])
        '''

    def compute_partials(self, inputs, J):
        J['wind_zvec','hub_height'] = np.ones(1)
        '''


        
class TurbineClass(ExplicitComponent):
    def setup(self):
        # parameters
        self.add_discrete_input('turbine_class', val='I', desc='IEC turbine class')
        self.add_input('V_mean_overwrite', val=0., desc='overwrite value for mean velocity for using user defined CDFs')

        # outputs should be constant
        self.add_output('V_mean', shape=1, units='m/s', desc='IEC mean wind speed for Rayleigh distribution')
        self.add_output('V_extreme1', shape=1, units='m/s', desc='IEC extreme wind speed at hub height')
        self.add_output('V_extreme50', shape=1, units='m/s', desc='IEC extreme wind speed at hub height')
        self.add_output('V_extreme_full', shape=2, units='m/s', desc='IEC extreme wind speed at hub height')
        
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        turbine_class = discrete_inputs['turbine_class'].upper()

        if turbine_class == 'I':
            Vref = 50.0
        elif turbine_class == 'II':
            Vref = 42.5
        elif turbine_class == 'III':
            Vref = 37.5
        elif turbine_class == 'IV':
            Vref = 30.0
        else:
            raise ValueError('turbine_class input must be I/II/III/IV')

        if inputs['V_mean_overwrite'] == 0.:
            outputs['V_mean'] = 0.2*Vref
        else:
            outputs['V_mean'] = inputs['V_mean_overwrite']
        outputs['V_extreme1'] = 0.8*Vref
        outputs['V_extreme50'] = 1.4*Vref
        outputs['V_extreme_full'][0] = 1.4*Vref # for extreme cases TODO: check if other way to do
        outputs['V_extreme_full'][1] = 1.4*Vref



class RotorGeometry(Group):
    def initialize(self):
        self.options.declare('RefBlade')
        self.options.declare('topLevelFlag',default=False)
        
        # Blade Cost Model Options
        self.options.declare('verbosity',        default=False)
        self.options.declare('tex_table',        default=False)
        self.options.declare('generate_plots',   default=False)
        self.options.declare('show_plots',       default=False)
        self.options.declare('show_warnings',    default=False)
        self.options.declare('discrete',         default=False)
        self.options.declare('user_update_routine', default=None)
        
    def setup(self):
        RefBlade = self.options['RefBlade']
        topLevelFlag   = self.options['topLevelFlag']
        NINPUT = len(RefBlade['ctrl_pts']['r_in'])
        NAF    = len(RefBlade['outer_shape_bem']['airfoil_position']['grid'])
        
        verbosity           = self.options['verbosity']
        tex_table           = self.options['tex_table']     
        generate_plots      = self.options['generate_plots']
        show_plots          = self.options['show_plots']
        show_warnings       = self.options['show_warnings']
        discrete            = self.options['discrete']
        user_update_routine = self.options['user_update_routine']
        
        # Independent variables that are unique to TowerSE
        if topLevelFlag:
            geomIndeps = IndepVarComp()
            geomIndeps.add_output('bladeLength', 0.0, units='m')
            geomIndeps.add_output('hubFraction', 0.0)
            geomIndeps.add_output('r_max_chord', 0.0)
            geomIndeps.add_output('chord_in', np.zeros(NINPUT),units='m')
            geomIndeps.add_output('theta_in', np.zeros(NINPUT), units='deg')
            geomIndeps.add_output('precurve_in', np.zeros(NINPUT), units='m')
            geomIndeps.add_output('presweep_in', np.zeros(NINPUT), units='m')
            # geomIndeps.add_output('precurveTip', 0.0, units='m')
            # geomIndeps.add_output('presweepTip', 0.0, units='m')
            geomIndeps.add_output('precone', 0.0, units='deg')
            geomIndeps.add_output('tilt', 0.0, units='deg')
            geomIndeps.add_output('yaw', 0.0, units='deg')
            geomIndeps.add_discrete_output('nBlades', 3)
            geomIndeps.add_discrete_output('downwind', False)
            geomIndeps.add_discrete_output('turbine_class', val='I', desc='IEC turbine class')
            geomIndeps.add_discrete_output('blade_in_overwrite', val={}, desc='IEC turbine class')
            geomIndeps.add_output('V_mean_overwrite', val=0.0, desc='optional overwrite value for mean velocity for using user defined CDFs')
            geomIndeps.add_output('airfoil_position', val=np.zeros(NAF))
            geomIndeps.add_output('sparT_in', val=np.zeros(NINPUT), units='m', desc='spar cap thickness parameters')
            geomIndeps.add_output('teT_in', val=np.zeros(NINPUT), units='m', desc='trailing-edge thickness parameters')
            # geomIndeps.add_output('leT_in', val=np.zeros(NINPUT), units='m', desc='leading-edge thickness parameters')
            self.add_subsystem('geomIndeps', geomIndeps, promotes=['*'])
            
        # --- Rotor Definition ---
        self.add_subsystem('loc', Location(), promotes=['*'])
        self.add_subsystem('turbineclass', TurbineClass(), promotes=['*'])
        #self.add_subsystem('spline0', BladeGeometry(RefBlade))
        self.add_subsystem('spline', BladeGeometry(RefBlade=RefBlade, 
                                               verbosity=verbosity,
                                               tex_table=tex_table,
                                               generate_plots=generate_plots,
                                               show_plots=show_plots,
                                               show_warnings =show_warnings ,
                                               discrete=discrete,
                                               user_update_routine=user_update_routine), promotes=['*'])
        self.add_subsystem('geom', CCBladeGeometry(NINPUT = NINPUT), promotes=['precone','precurve_in', 'presweep_in',
                                                                               'precurveTip','presweepTip','R','Rtip'])

        
def Init_RotorGeometry_wRefBlade(rotor, blade):
    rotor['precone']          = blade['config']['cone_angle']
    rotor['bladeLength']      = blade['ctrl_pts']['bladeLength'] #61.5  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
    rotor['r_max_chord']      = blade['ctrl_pts']['r_max_chord']  # 0.23577 #(Float): location of max chord on unit radius
    rotor['chord_in']         = np.array(blade['ctrl_pts']['chord_in']) # np.array([3.2612, 4.3254, 4.5709, 3.7355, 2.69923333, 1.4621])  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
    rotor['theta_in']         = np.array(blade['ctrl_pts']['theta_in']) # np.array([0.0, 13.2783, 12.30514836,  6.95106536,  2.72696309, -0.0878099]) # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
    rotor['precurve_in']      = np.array(blade['ctrl_pts']['precurve_in']) #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
    rotor['presweep_in']      = np.array(blade['ctrl_pts']['presweep_in']) #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
    rotor['sparT_in']         = np.array(blade['ctrl_pts']['sparT_in']) # np.array([0.0, 0.05, 0.047754, 0.045376, 0.031085, 0.0061398])  # (Array, m): spar cap thickness parameters
    rotor['teT_in']           = np.array(blade['ctrl_pts']['teT_in']) # np.array([0.0, 0.1, 0.09569, 0.06569, 0.02569, 0.00569])  # (Array, m): trailing-edge thickness parameters
    # if 'le_var' in blade['precomp']['le_var']:
    #     rotor['leT_in']       = np.array(blade['ctrl_pts']['leT_in']) ## (Array, m): leading-edge thickness parameters
    rotor['airfoil_position'] = np.array(blade['outer_shape_bem']['airfoil_position']['grid'])
    rotor['hubFraction']      = blade['config']['hubD']/2./blade['pf']['r'][-1] #0.025  # (Float): hub location as fraction of radius
    rotor['hub_height']       = blade['config']['hub_height']  # (Float, m): hub height
    rotor['turbine_class']    = blade['config']['turbine_class'].upper() #TURBINE_CLASS['I']  # (Enum): IEC turbine class

    return rotor


if __name__ == "__main__":

    # Turbine Ontology input
    # fname_input = "turbine_inputs/nrel5mw_mod_update.yaml"
    fname_input = "/mnt/c/Users/egaertne/WISDEM2/wisdem/Design_Opt/nrel15mw/inputs/NREL15MW_prelim_v3.0.yaml"

    # Initialize blade design
    refBlade = ReferenceBlade()
    refBlade.verbose  = True
    refBlade.NINPUT   = NINPUT
    refBlade.NPTS     = 50
    refBlade.validate = False
    refBlade.fname_schema = "turbine_inputs/IEAontology_schema.yaml"

    refBlade.spar_var = ['Spar_Cap_SS', 'Spar_Cap_PS']
    refBlade.te_var   = 'TE_reinforcement'
    # refBlade.le_var       = 'le_reinf'
    blade = refBlade.initialize(fname_input)

    # setup
    rotor = Problem()
    rotor.model = RotorGeometry(RefBlade=blade, topLevelFlag=True)
    rotor.setup()
    rotor = Init_RotorGeometry_wRefBlade(rotor, blade)
    rotor.run_driver()

    print(rotor['total_blade_cost'])
